from typing import List
from slither.utils.output import Output
from slither.core.declarations import Function
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification


class InconsistentNonreentrant(AbstractDetector):
    """
    Sees if contract non-view functions do not have "nonreentrant" modifier while other functions have it.
    """

    ARGUMENT = 'pess-inconsistent-nonreentrant' # slither will launch the detector with slither.py --detect inconsistent-nonreentrant
    HELP = 'function ... (), function ... () nonReentrant'
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/inconsistent_nonreentrant.md'
    WIKI_TITLE = 'Inconsistent nonreentrant'
    WIKI_DESCRIPTION = "If non-reentrant modifier is used then for security reasons it should be used in all non-veiw functions"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Make sure that all non-view functions use non-reentrant modifier'


    def _has_visibility_or_view(self, fun: Function) -> bool:
        """Function which checks if function is view or private/internal"""
        if fun.view:
            return True
        if fun.visibility in {'private', 'internal'}:
            return True

    def _has_modifiers(self, fun: Function) -> bool:
        """Function which checks modifiers on functions"""
        for m in fun.modifiers:
            if str(m.name) in 'nonReentrant':
                return True

    def _not_empty_function(self, fun: Function) -> bool:
        """Function which checks if given function is not empty"""
        if len(fun.nodes) != 0:
            return True

    def _validate_detection(self, contract_functions, nonreentrant_count) -> bool:
        """Function which validates the detection by comparing the number of functions and functions with nonReentrant"""
        contract_functions_length = len(contract_functions)
        nonreentrant_count_length = len(nonreentrant_count)
        if contract_functions_length == nonreentrant_count_length or nonreentrant_count_length == 0:
            return True

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        contract_functions = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                if f.name != "constructor" and not self._has_visibility_or_view(f) and self._not_empty_function(f):    # contstructor function is not checked
                    contract_functions.append(f)
                    if not self._has_modifiers(f):
                        res.append(self.generate_result([
                            "Function", ' ',
                            f, ' is a non-view function without nonReentrant modifier'
                            '\n']))
            invalid_detection = self._validate_detection(contract_functions, res)
            if invalid_detection:
                res.clear()
            return res



import re
from collections import defaultdict
from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function


class MagicNumber(AbstractDetector):
    """
    Shows int/uint values which are not assigned to variables.
    """

    ARGUMENT = "pess-magic-number"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "int/uint values except 0, 1, 2, 1000 and 1e18"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/pessimistic-io/slitherin/blob/master/docs/magic_number.md"
    )
    WIKI_TITLE = "Magic Number"
    WIKI_DESCRIPTION = "Values should be assigned to variables"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Assign values to variables"

    EXCEPTION = {"0", "1", "2", "1000", "1e18"}
    used_count = defaultdict(lambda: {"count": 0, "nodes": []})

    def _check_if_pow_10(self, str: str) -> bool:
        reg = re.fullmatch(r"^10*$|^10*e\d+$", str)  # 1(0..) or 1(0..)eX
        if reg:
            return True
        return False

    def _getLiterals(self, fun: Function) -> None:
        """Get numbers except those which are in EXCEPTION{}"""

        if (
            fun.name != "slitherConstructorConstantVariables"
        ):  # removing values assigned to constant variables
            for n in fun.nodes:
                nodeString = str(n)
                lit = re.search(
                    r"\W\d+e\d+|\W\d+[_*\d]*", nodeString
                )  # regular expression to get numbers
                if lit:
                    magic_number = lit[0]
                    index = (
                        len(magic_number) - 1
                    )  # get length of the found number and remove the space in the beginning
                    if magic_number[
                        -index:
                    ] not in self.EXCEPTION and not self._check_if_pow_10(
                        magic_number[-index:]
                    ):
                        result_number = magic_number[-index:]
                        self.used_count[result_number]["count"] += 1
                        self.used_count[result_number]["nodes"].append(n)

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        all_used_counts = []
        for contract in self.compilation_unit.contracts_derived:
            self.used_count = defaultdict(lambda: {"count": 0, "nodes": []})
            for f in contract.functions:
                self._getLiterals(f)
            all_used_counts.append(self.used_count)

        for contract_used_count in all_used_counts:
            for num, data in contract_used_count.items():
                if num:
                    if data["count"] < 2:
                        continue
                    info = [f"Magic number {num} is used multiple times in:\n"]
                    for n in data["nodes"]:
                        info += ["\t", n, "\n"]
                    res.append(self.generate_result(info))
        return res



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.core.solidity_types.elementary_type import Int, Uint
from  slither.core.variables.local_variable import LocalVariable
import slither.slithir.operations as ops
import slither.slithir.variables as vrs
from slither.core.cfg.node import NodeType
import typing as tp

INT_TYPES = [*Int, *Uint]

class PotentialArithmOverflow(AbstractDetector):
    """
    Detects expressions where overflow may occur, but no overflow is expected
    """

    ARGUMENT = "pess-potential-arithmetic-overflow"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Add explicit type casting in temporary expressions during the evaluation of arithmetic expressions in case the final type is larger than the intermediate one."
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/pessimistic-io/slitherin/blob/master/docs/potential_arith_overflow.md"
    )
    WIKI_TITLE = "Potential Arithmetic Overflow"
    WIKI_DESCRIPTION = "The detector sees if there are assignments/returns that calculate some arithmetic expressions and if some intermediate calculations contain a type that is lower than the expected result. Such behavior may lead to unexpected overflow/underflow, e.g., trying to assign the multiplication of two `uint48` variables to `uint256` would look like `uint48 * uint48` and it may overflow (however, the final type would fit such multiplication)."
    WIKI_EXPLOIT_SCENARIO = "A transaction will revert with overflow, but the expected behavior that it should not."
    WIKI_RECOMMENDATION = "Use explicit type casting in sub expressions when the assigment to a larger type is performed."
    
    def _has_op_overflowing_sub_expression(self, op: ops.Operation, high_level_bits_needed: tp.Optional[int]) -> (bool, list, str):
        if isinstance(op, ops.Unary):
            (is_err, nodes, sub_expr) = self._has_op_overflowing_sub_expression(op.rvalue, high_level_bits_needed) # ! and ~ do not affect result
            return (is_err, nodes, f"{op.type_str} {sub_expr}")
        elif isinstance(op, ops.Binary):
            result_type = str(op.lvalue.type)
            [l_check_res, l_check_list, l_sub_expr] = self._has_op_overflowing_sub_expression(op.variable_left, high_level_bits_needed)
            [r_check_res, r_check_list, r_sub_expr] = self._has_op_overflowing_sub_expression(op.variable_right, high_level_bits_needed)
                
            cur_expr = f"{l_sub_expr} {op._type.value} {r_sub_expr}"
            if op._type is ops.BinaryType.ADDITION or op._type is ops.BinaryType.MULTIPLICATION or op._type is ops.BinaryType.POWER:
                result_bits_cut = result_type.removeprefix("uint").removeprefix("int")
                result_bits = int(256 if not result_bits_cut else result_bits_cut)
                if high_level_bits_needed and high_level_bits_needed > result_bits: # result expects bigger type, but expression returns lower and overflow may occur here
                    return (True, [(cur_expr, result_type), *l_check_list, *r_check_list], cur_expr)
                else:
                    return (l_check_res or r_check_res, [*l_check_list, *r_check_list], cur_expr)
            else: # @todo currently we do not check other operations, return as everything is ok
                return (l_check_res or r_check_res, [*l_check_list, *r_check_list], cur_expr)
        elif isinstance(op, ops.OperationWithLValue):
            return (False, [], f"{op.lvalue}")
        elif isinstance(op, vrs.Constant):
            return (False, [], f"{op}")
        elif isinstance(op, LocalVariable):
            return (False, [], f"{op}")
        elif isinstance(op, vrs.state_variable.StateVariable):
            return (False, [], f"{op}")
        elif isinstance(op, vrs.TemporaryVariable):
            return (False, [], f"...")
        elif isinstance(op, ops.Return):
            return (False, [], f"{op}")
        elif isinstance(op, vrs.ReferenceVariable):
            return (False, [], f"{op}")
        else:
            return (False, [], f"{op}")

    def _find_vulnerable_expressions(self, fun: Function) -> list:
        final_results = []
        is_active_assembly = False
        for node in fun.nodes:
            if node.type is NodeType.ASSEMBLY:
                is_active_assembly = True
            elif node.type is NodeType.ENDASSEMBLY:
                is_active_assembly = False
            if (node.type is NodeType.VARIABLE or node.type is NodeType.EXPRESSION or node.type is NodeType.RETURN) and not is_active_assembly:
                irs = node.irs
                if len(irs) > 0 and isinstance(irs[-1], ops.Assignment) and str(irs[-1].lvalue._type) in INT_TYPES:
                    expected_bits_cut = str(irs[-1].lvalue._type).removeprefix("uint").removeprefix("int")
                    expected_bits = int(256 if not expected_bits_cut else expected_bits_cut)
                    has_problems = False
                    errors = []
                    for x in irs[:-1]:
                        (response_res, response_errors, _) = self._has_op_overflowing_sub_expression(x, expected_bits)
                        has_problems |= response_res
                        errors.extend(response_errors)
                    if has_problems:
                        final_results.append((node, str(irs[-1].lvalue._type), errors))
                if len(irs) > 0 and isinstance(irs[-1], ops.Return) and len(fun.return_type) == 1 and str(fun.return_type[0]) in INT_TYPES: # @todo currently works only with single returns
                    expected_bits_cut = str(fun.return_type[0]).removeprefix("uint").removeprefix("int")
                    expected_bits = int(256 if not expected_bits_cut else expected_bits_cut)
                    has_problems = False
                    errors = []
                    for x in irs[:-1]:
                        (response_res, response_errors, _) = self._has_op_overflowing_sub_expression(x, expected_bits)
                        has_problems |= response_res
                        errors.extend(response_errors)
                    if has_problems:
                        final_results.append((node, str(fun.return_type[0]), errors))

        return final_results
    
    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if contract.is_interface:
                continue
            for f in contract.functions_and_modifiers_declared:
                vulnerable_expressions = self._find_vulnerable_expressions(f)
                if vulnerable_expressions:
                    info = [f, f" contains integer variables whose type is larger than the type of one of its intermediate expressions. Consider casting sub expressions explicitly as they might lead to unexpected overflow:\n"]
                    for (node, node_final_type, op_with_ret_type) in vulnerable_expressions:
                        info += ["\tIn `", node, "` intermidiate expressions returns type of lower order:", "\n"]
                        for (op, op_ret_type) in op_with_ret_type:
                            info += ["\t\t`", str(op), f"` returns {op_ret_type}, but the type of the resulting expression is {node_final_type}.", "\n"]
                    res.append(self.generate_result(info))
        return res



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import LowLevelCall
from slither.core.declarations import Function
from slither.utils.output import Output
from pathlib import Path
from typing import List

OZ_EXCLUDED_CONTRACTS_SUFFIXES: List[str] = [
    "contracts/utils/Address.sol",
    "contracts-upgradeable/utils/AddressUpgradeable.sol",
    "contracts/utils/AddressUpgradeable.sol",
    "contracts/token/ERC20/utils/SafeERC20.sol"
]

UNISWAP_EXCLUDED_CONTRACTS_SUFFIXES: List[str] = [
    "lib/contracts/libraries/TransferHelper.sol"
]

BALANCER_EXCLUDED_CONTRACTS_SUFFIXES: List[str] = [
    "v2-solidity-utils/contracts/openzeppelin/SafeERC20.sol"
]
class CallForwardToProtected(AbstractDetector):
    """
    Shows cases when contract calls custom addresses and has contract interactions through access control
    """

    ARGUMENT = 'pess-call-forward-to-protected' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'Contract might have a call to a custom address'
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/call_forward_to_protected.md'
    WIKI_TITLE = 'Call Forward To Protected'
    WIKI_DESCRIPTION = "No calls to custom addresses and contract interactions through access control"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Do not let calls to project contracts or do not give rights to a contract which can perform calls'

    
    def _contains_low_level_calls(self, node) -> bool:
        return any(isinstance(ir, LowLevelCall) for ir in node.irs)

    def _detect_low_level_custom_address_call(self, fun: Function) -> bool:
        address_parameters = [
            parameter for parameter in fun.parameters if str(parameter.type) == "address" and parameter.name
        ]
        for node in fun.nodes:
            if self._contains_low_level_calls(node):
                for address_parameter in address_parameters:
                    if str(address_parameter) in str(node):
                        return True
        return False
    
    def _pess_is_excluded_from_detector(self, contract: "Contract") -> bool:
        path = Path(contract.source_mapping.filename.absolute).parts
        is_zep = "openzeppelin-solidity" in path or \
                    ("@openzeppelin" in path and path[path.index("@openzeppelin") + 1] == "contracts") or \
                    "openzeppelin-contracts-upgradeable" in path or \
                    ("@openzeppelin-contracts-upgradeable" in path and \
                        (path[path.index("@openzeppelin-contracts-upgradeable") + 1] == "contracts" or \
                         path[path.index("@openzeppelin-contracts-upgradeable") + 1] == "contracts-upgradeable"))

        if is_zep:
            return any(map(lambda suffix: contract.source_mapping.filename.absolute.endswith(suffix), OZ_EXCLUDED_CONTRACTS_SUFFIXES)) 
        elif "@uniswap" in path:
            return any(map(lambda suffix: contract.source_mapping.filename.absolute.endswith(suffix), UNISWAP_EXCLUDED_CONTRACTS_SUFFIXES)) 
        elif "@balancer-labs" in path: 
            return any(map(lambda suffix: contract.source_mapping.filename.absolute.endswith(suffix), BALANCER_EXCLUDED_CONTRACTS_SUFFIXES)) 
        return False

    def _detect(self) -> List[Output]:
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if self._pess_is_excluded_from_detector(contract):
                continue
            for f in contract.functions:
                x = self._detect_low_level_custom_address_call(f)
                if x:
                    res.append(self.generate_result([
                            "Function", ' ',
                            f, ' contains a low level call to a custom address',
                            '\n']))

        return res



from logging import Logger
import sys
import os
import json
import copy
from typing import List, Optional
from slither import Slither
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function, Contract
from slither.core.expressions.type_conversion import TypeConversion
from slither.core.cfg.node import Node
from slither.core.variables.local_variable import LocalVariable
from slither.core.solidity_types.array_type import ArrayType
from slither.slithir.operations import InternalCall, HighLevelCall
from slither.slithir.operations.assignment import Assignment
from slither.analyses.data_dependency.data_dependency import is_dependent


class Context:
    def __init__(
        self,
        c: Contract,
        entry_point: Function,
        func: Function,
        args_tainted_mask: List[bool],
        verbose: bool,
    ):
        self.visited = []
        self.contract = c
        self.entry_point = entry_point
        self.hit_function = None
        self.tainted_path = []
        self.param_name = None
        self.function = func
        self.args_tainted_mask = args_tainted_mask  # TODO remove as this is redundant (covered by entrypoint_params_index_dependency)
        self.entrypoint_params_index_dependency: List[List[int]] = [
            [i] for i in range(len(entry_point.parameters))
        ]
        self.call_chain = [entry_point.name]
        self.hits = []
        self.verb = verbose

    def make_copy(self):  # to avoid passing by reference in internal calls inspections
        cp = copy.copy(self)
        cp.call_chain = copy.copy(self.call_chain)
        cp.entrypoint_params_index_dependency = copy.copy(
            self.entrypoint_params_index_dependency
        )
        return cp

    def __str__(self):  # debug
        return (
            "entry_point"
            + str(self.entry_point)
            + "hit_function"
            + str(self.hit_function)
            + "param_name"
            + str(self.param_name)
        )


class Hit:
    def __init__(
        self,
        entry_point: Function,
        hit_function: Function,
        param_name,
        call_chain: List[str],
        entry_params_used: List[str],
        swap_func: str,
    ):
        self.entry_point = entry_point
        self.hit_function = hit_function
        self.param_name = param_name
        self.call_chain = call_chain
        self.entry_params_used = entry_params_used
        self.swap_func = swap_func


swap_functions = [
    "swapExactTokensForTokens",
    "swapTokensForExactTokens",
    "swapExactTokensForTokensSupportingFeeOnTransferTokens",
    "swapExactETHForTokensSupportingFeeOnTransferTokens",
    "swapExactTokensForETHSupportingFeeOnTransferTokens",
    "swapETHForExactTokens",
    "swapExactTokensForETH",
    "swapTokensForExactETH",
    "swapExactETHForTokens",
    "getAmountsOut",
    "getAmountsIn",
]
swap_route_arg_pos = [2, 2, 2, 1, 2, 1, 2, 2, 1, 1, 1]

# widespread false positives
banned_funcs = ["ensure", "constructor"]
# partial (LIKE %x%)
banned_funcs_partial = []
banned_contracts_partial = ["swap", "dex", "router", "pancake", "buysell"]


def check_contract(c: Contract) -> List[Function]:
    results_raw: List[Function] = []

    if len(swap_functions) != len(swap_route_arg_pos):
        print("ERROR - Array lengths not matching")
        return results_raw
    # skip test contracts
    if any(x in c.name for x in ["Test", "Mock"]) or any(
        x in c.name.lower() for x in banned_contracts_partial
    ):
        return results_raw

    # to filter constructors for old solidity versions
    inherited_contracts = get_inherited_contracts(c) + [c.name]

    # filter unwanted funcs (also checked later recursively for nested calls)
    to_inspect = (
        x
        for x in c.functions
        if (
            # filter banned functions and old solc constructors
            x.name not in (banned_funcs + inherited_contracts)
            and not any(y in x.name for y in banned_funcs_partial)
            and not x.pure
            and not x.view
            and x.visibility in ["public", "external"]
            and x.is_implemented  # internal functions inspected through internal calls to keep the flow consistent
        )
    )

    for f in to_inspect:
        verb = False
        if f.name == "-":
            verb = True
        ctx = Context(c, f, f, [True for i in range(len(f.parameters))], verb)
        check_function(f.entry_point, ctx)
        if len(ctx.hits):
            results_raw += ctx.hits
    return results_raw


def check_function(node: Optional[Node], ctx: Context):
    if node is None:
        return False

    if ctx.function.name in banned_funcs or any(
        x in ctx.function.name for x in banned_funcs_partial
    ):
        return False

    if is_modifier_protected(ctx.function):  # TODO enhance
        return False

    if node in ctx.visited:
        return False
    ctx.visited.append(node)

    # debug
    if ctx.verb:
        print(str(node))
        for ir in node.irs:
            print("    " + str(ir) + "    " + str(type(ir)))

    # check if the input parameters are checked, and remove taint flag if so
    if (
        node.contains_if()
        or node.contains_require_or_assert()
        and "address(0)" not in str(node)
    ):
        for i in range(len(ctx.function.parameters)):
            param = ctx.function.parameters[i]
            if isinstance(param, LocalVariable) and not isinstance(
                param.type, ArrayType
            ):  # only look for checks on single addresses, not on path (which most likely will be length)
                if param in node.local_variables_read:
                    ctx.args_tainted_mask[i] = False
                    ctx.entrypoint_params_index_dependency[i] = False

    for ir in node.irs:
        # check for external calls
        if isinstance(ir, HighLevelCall):
            for i in range(len(swap_functions)):  # check if
                if (
                    hasattr(ir.function, "name")
                    and ir.function.name == swap_functions[i]
                    and len(ir.arguments) >= swap_route_arg_pos[i] + 1
                ):
                    path_argument = ir.arguments[swap_route_arg_pos[i]]
                    tainted_indexes = is_dependent_on_any_tainted(
                        path_argument,
                        ctx.function.parameters,
                        ctx.args_tainted_mask,
                        ctx.function,
                    )
                    if len(tainted_indexes):
                        if not (
                            ctx.function.name == ctx.entry_point.name
                            and ctx.function.name in swap_functions
                        ):
                            _ctx = ctx.make_copy()
                            # list entry parameters used to compute the route
                            entry_params_used = []
                            for (
                                ti
                            ) in (
                                tainted_indexes
                            ):  # indici dei parametri della funzione corrente, da cui dipende la path
                                for pi in ctx.entrypoint_params_index_dependency[
                                    ti
                                ]:  # indici dei parametri dell'entrypoint
                                    entry_params_used.append(
                                        ctx.entry_point.parameters[pi].name
                                    )
                            ctx.hits.append(
                                Hit(
                                    _ctx.entry_point,
                                    _ctx.function,
                                    path_argument.name,
                                    _ctx.call_chain,
                                    entry_params_used,
                                    swap_functions[i],
                                )
                            )

        # internal call to inspect
        elif isinstance(ir, InternalCall):
            # prepare new context for the callee
            ctx_callee = ctx.make_copy()
            ctx_callee.call_chain.append(ir.function.name)
            # map tainted bool mask from caller parameters to callee arguments
            taint_mask = []
            ctx_callee.entrypoint_params_index_dependency = [
                [] for i in range(len(ir.arguments))
            ]  # empty out to fill in the coming loop
            for i in range(len(ir.arguments)):
                carg = ir.arguments[i]
                tainted_indexes = is_dependent_on_any_tainted(
                    carg, ctx.function.parameters, ctx.args_tainted_mask, ctx.function
                )
                taint_mask.append(True if len(tainted_indexes) else False)
                for ti in tainted_indexes:
                    for epi in ctx.entrypoint_params_index_dependency[ti]:
                        ctx_callee.entrypoint_params_index_dependency[i].append(epi)

            ctx_callee.args_tainted_mask = taint_mask
            ctx_callee.function = ir.function
            check_function(ir.function.entry_point, ctx_callee)
            if len(ctx_callee.hits) > len(ctx.hits):  # callee has new hits
                for i in range(len(ctx_callee.hits), len(ctx.hits)):
                    ctx.hits.append(ctx_callee.hits[i])

    # sons inspection
    ret = False
    for son in node.sons:
        ret = ret or check_function(
            son, ctx
        )  # ctx by reference to get it populated with flags
    return ret


def is_dependent_on_any_tainted(
    checked_var: LocalVariable,
    vars_list: List[LocalVariable],
    tainted_mask: List[bool],
    context,
) -> List[int]:
    ret: List[int] = []
    if not type(checked_var) is list:  # can be list eg foo([bar, 0], par)
        checked_var = [checked_var]
    for i in range(len(checked_var)):
        if not isinstance(checked_var[i], LocalVariable):
            continue
        if not len(vars_list) == len(tainted_mask):
            print("ERROR: Mismatch array length vars_list vs tainted_mask")
            return []
        for j in range(len(vars_list)):
            if not tainted_mask[j]:
                continue
            if is_dependent(checked_var[i], vars_list[j], context):
                ret.append(j)
    return ret


def get_inherited_contracts(contract: Contract) -> List[str]:
    return list(x.name for x in contract.inheritance)


def is_modifier_protected(func: Function) -> bool:
    bannedPattern = [
        "admin",
        "owner",
        "role",
        "only",
        "permission",
        "initializ",
        "auth",
    ]
    for mod in func.modifiers:
        if any(x in mod.name.lower() for x in bannedPattern):
            return True
    return False


class UniswapV2(AbstractDetector):
    """
    Checks code for UniswapV2 integration vulnerabilities
    """

    ARGUMENT = "pess-uni-v2"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Detector is based on UniswapV2 checklist"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/integration_uniswapV2.md"
    WIKI_TITLE = "UniswapV2 Integration"
    WIKI_DESCRIPTION = (
        "UniswapV2 integration vulnerabilities must not be present in the codebase"
    )
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Follow UniswapV2 integration checklist"

    def __init__(
        self, compilation_unit: SlitherCompilationUnit, slither: Slither, logger: Logger
    ) -> None:
        super().__init__(compilation_unit, slither, logger)
        absolute_path = os.path.dirname(__file__)
        relative_path = "../utils/deflat_tokens.json"
        full_path = os.path.join(absolute_path, relative_path)
        fileJson = open(full_path)
        data = json.load(fileJson)
        self.token_objects = data["objects"]
        fileJson.close()

    def _pair_balance_used(self, fun: Function) -> bool:
        """Checks if a function a uses pair balance"""
        mb_pair_vars = []
        for node in fun.nodes:
            if (
                node.state_variables_read or node.local_variables_read
            ):  # Checks if IUniswapV2Pair type is in local/storage vars
                for node_var in node.state_variables_read:
                    if str(node_var.type) == "IUniswapV2Pair":
                        mb_pair_vars.append(str(node_var))
                for node_var in node.local_variables_read:
                    if str(node_var.type) == "IUniswapV2Pair":
                        mb_pair_vars.append(str(node_var))
            for (
                external_call
            ) in (
                node.external_calls_as_expressions
            ):  # Checks external calls for expressions with balanceOf
                if "balanceOf" in str(external_call):
                    for call_arg in external_call.arguments:
                        if isinstance(call_arg, TypeConversion):
                            if str(  # Checks if IUniswapV2Pair is used in balanceOf
                                call_arg.expression
                            ) in mb_pair_vars or "IUniswapV2Pair" in str(
                                call_arg.expression
                            ):
                                return True
        return False

    def _pair_reserve_used(self, fun: Function) -> bool:
        """Checks if a function uses getReserves function of the Pair contract"""
        for external_call in fun.external_calls_as_expressions:
            if (
                "getReserves" in str(external_call)
                and "tuple" in external_call.type_call
                and "function () view external returns (uint112,uint112,uint32)"
                in external_call.called.type
            ):
                return True
        return False

    def _pair_used(self, contract: Contract) -> bool:
        """Checks if a Pair contract is used"""
        for var in contract.variables:  # Looking for a IUniswapV2Pair type var
            if str(var.type) == "IUniswapV2Pair":
                return True
        for function in contract.functions:
            for node in function.nodes:
                if "IUniswapV2Pair" in str(
                    node
                ):  # Looking for IUniswapV2Pair in functions
                    return True
        return False

    def _minReturn_zero(self, fun: Function) -> bool:
        """Checks if a minReturn parameter equals zero"""
        calls_with_0_pos_argument = (
            {  # Functions where minReturn is an argument at position 0
                "swapExactETHForTokens",
                "swapExactETHForTokensSupportingFeeOnTransferTokens",
            }
        )
        calls_with_1_pos_argument = (
            {  # Functions where minReturn is an argument at position 1
                "swapExactTokensForTokens",
                "swapExactTokensForETH",
                "swapExactTokensForTokensSupportingFeeOnTransferTokens",
                "swapExactTokensForETHSupportingFeeOnTransferTokens",
            }
        )
        amountOutMin = ""
        for (
            external_call
        ) in (
            fun.external_calls_as_expressions
        ):  # Looking for a variable in the needed external call at positions 0 and 1
            if [
                call
                for call in calls_with_0_pos_argument
                if (call in str(external_call))
            ] and external_call.arguments:
                amountOutMin = external_call.arguments[0]
            if [
                call
                for call in calls_with_1_pos_argument
                if (call in str(external_call))
            ] and len(external_call.arguments) > 1:
                amountOutMin = external_call.arguments[1]
        if str(amountOutMin) == "0":  # If argument is set directly to 0 return True
            return True
        elif amountOutMin:  # Looking for 0 assignments to a variable found above
            for node in fun.nodes:
                for ir in node.irs:
                    if isinstance(ir, Assignment):
                        if (
                            str(ir.lvalue) == str(amountOutMin)
                            and str(ir.rvalue) == "0"
                        ):
                            return True
        return False

    def _maxReturn_max(self, fun: Function) -> bool:
        """Checks if a maxReturn parameter equals max"""
        calls_with_1_pos_argument = (
            {  # Functions where minReturn is an argument at position 1
                "swapTokensForExactTokens",
                "swapTokensForExactETH",
            }
        )
        amountInMax = ""
        tmp = ""
        for (
            external_call
        ) in (
            fun.external_calls_as_expressions
        ):  # Looking for a variable in the needed external call at position 1
            if [
                call
                for call in calls_with_1_pos_argument
                if (call in str(external_call))
            ] and len(external_call.arguments) > 1:
                amountInMax = external_call.arguments[1]
        if (
            str(amountInMax) == "type()(uint256).max"
        ):  # If argument is set directly to 0 return True
            return True
        elif amountInMax:  # Looking for max assignments to a variable found above
            for node in fun.nodes:
                for ir in node.irs:
                    if isinstance(ir, Assignment):
                        if (
                            "TMP_" in str(ir.lvalue)
                            and str(ir.rvalue)
                            == "115792089237316195423570985008687907853269984665640564039457584007913129639935"
                        ):
                            tmp = str(ir.lvalue)
                        if str(amountInMax) == str(ir.lvalue) and tmp == str(ir.rvalue):
                            return True
        return False

    def _has_bad_token(self, fun: Function) -> bool:
        """Checks if deflationary or rebase tokens are used"""

        for (
            section
        ) in self.token_objects:  # Looks for token addresses from json in nodes
            address = section["address"]
            for node in fun.nodes:
                if address in str(node):
                    return True

        return False

    def _contract_uses_uniswap(self, c: Contract) -> bool:
        for (
            fcontract,
            _,
        ) in c.all_high_level_calls:
            if "uniswapv2" in fcontract.name.lower():
                return True

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        results_raw: List[Hit] = []
        results: List[Output] = []

        for contract in self.compilation_unit.contracts_derived:
            contains_uni_v2 = self._contract_uses_uniswap(contract)
            if (
                not contains_uni_v2 and not "pess-uni-v2" in sys.argv
            ):  # launch only if detector in terminal is present or there is dependency
                continue
            results_raw += check_contract(contract)

            for f in results_raw:
                info = [
                    f"Tainted route '",
                    f.param_name,
                    "' for function '",
                    f.swap_func,
                    "'\n in function '",
                    f.hit_function,
                    "'\n from entry point '",
                    f.entry_point,
                    "'\n Call chain (one of the possible ones): ",
                    str(f.call_chain),
                    "\n Entrypoint params used: ",
                    str(f.entry_params_used),
                    "\n\n",
                ]
                res.append(self.generate_result(info))
            pair_used = self._pair_used(contract)
            for f in contract.functions:
                pair_balance_used = self._pair_balance_used(f)
                pair_reserve_used = self._pair_reserve_used(f)
                minReturn_zero = self._minReturn_zero(f)
                maxReturn_max = self._maxReturn_max(f)
                has_bad_token = self._has_bad_token(f)
                if pair_balance_used:
                    res.append(
                        self.generate_result(
                            ["Function ", f, " uses pair balance." "\n"]
                        )
                    )
                if pair_reserve_used:
                    res.append(
                        self.generate_result(
                            ["Function ", f, " uses pair reserves." "\n"]
                        )
                    )
                if pair_used:
                    res.append(
                        self.generate_result(
                            [
                                "Contract ",
                                contract,
                                " uses pair contract directly." "\n",
                            ]
                        )
                    )
                if minReturn_zero:
                    res.append(
                        self.generate_result(
                            [
                                "Function ",
                                f,
                                ' has a call of a swap with "min" parameter. Min parameter must not be 0'
                                "\n",
                            ]
                        )
                    )
                if maxReturn_max:
                    res.append(
                        self.generate_result(
                            [
                                "Function ",
                                f,
                                ' has a call of a swap with "max" parameter. Max parameter must not be infinity'
                                "\n",
                            ]
                        )
                    )
                if has_bad_token:
                    res.append(
                        self.generate_result(
                            [
                                "Function ",
                                f,
                                " uses deflationary or rebase token" "\n",
                            ]
                        )
                    )
        return res



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.slithir.operations.event_call import EventCall


class EventSetter(AbstractDetector):
    """
    Sees if contract setters do not emit events
    """

    ARGUMENT = 'pess-event-setter' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'Contract function does not emit event after the value is set'
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/event_setter.md'
    WIKI_TITLE = 'Missing Event Setter'
    WIKI_DESCRIPTION = "Setter-functions must emit events"
    WIKI_EXPLOIT_SCENARIO = 'N/A'
    WIKI_RECOMMENDATION = 'Emit events in setter functions'


    def _emits_event(self, fun: Function) -> bool: 
        """Checks if function has multiple storage read"""
        if isinstance(fun, Function):   # check for a correct function type
            if any(ir for node in fun.nodes for ir in node.irs if isinstance(ir, EventCall)):
                return True
            return False

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if not contract.is_interface:
                for f in contract.functions_and_modifiers_declared:
                    if f.name.startswith("set"):
                        x = self._emits_event(f)
                        if not x:
                            res.append(self.generate_result([
                                "Setter function ",
                                f, ' does not emit an event'
                                '\n']))
        return res



from slither.core.cfg.node import NodeType

from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification


class UnprotectedSetter(AbstractDetector):
    """
    Sees if contract contains a setter, that changes contract paramater without modifier protection or access control inside the function
    """

    ARGUMENT = "pess-unprotected-setter"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Contract critical storage parameters might be changed by anyone"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/unprotected_setter.md"
    WIKI_TITLE = "Unprotected Setter"
    WIKI_DESCRIPTION = "Usually all setters must be protected with access control"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Add access control"

    def is_setter(self, fun, params=None):
        if not params:
            params = fun.parameters
        for n in fun.nodes:
            if n.type == NodeType.EXPRESSION:
                for v in n.state_variables_written:
                    lr = str(n.expression).split(" = ")
                    if len(lr) > 1:
                        left = lr[0]
                        right = lr[1]
                        for p in params:
                            if p.name:
                                if "." in left:
                                    continue
                                if "[" in left:
                                    continue
                                if right == str(p):
                                    return left
        return None

    def has_access_control(self, fun):
        for m in fun.modifiers:
            if m.name in ["initializer", "onlyOwner"] or m.name.startswith("only"):
                return True
        if fun.visibility in ["internal", "private"]:
            return True
        return fun.is_protected()

    def _detect(self):
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if contract.is_interface:
                continue
            for f in contract.functions:
                if not self.has_access_control(f):
                    x = self.is_setter(f)
                    if x != None:
                        res.append(
                            self.generate_result(
                                [
                                    "Function",
                                    " ",
                                    f,
                                    " is a non-protected setter ",
                                    x,
                                    " is written" "\n",
                                ]
                            )
                        )
        return res



import os
from typing import List
from slither.utils.output import Output
from slither.core.cfg.node import Node
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.slithir.operations.event_call import EventCall
from slither.analyses.data_dependency.data_dependency import is_dependent
from slither.slithir.operations import SolidityCall
from slither.core.declarations.solidity_variables import (
    SolidityVariableComposed,
    SolidityFunction,
)

from ...consts import ARBITRUM_KEY


class ArbitrumBlockNumberTimestamp(AbstractDetector):
    """
    Sees if `block.number` or `block.timtestamp` is used inside an Arbitrum contract
    """

    ARGUMENT = "pess-arb-block-number-timestamp"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "`block.number` or `block.timtestamp` is used inside an Arbitrum contract"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/arb_block_number_timestamp.md"
    WIKI_TITLE = (
        "Usage of `block.number` or `block.timtestamp` inside the Arbitrum contract"
    )
    WIKI_DESCRIPTION = "# Arbitrum block.number/block.timestamp"
    WIKI_EXPLOIT_SCENARIO = "N/A"
    WIKI_RECOMMENDATION = "Look at docs for details"

    def _find_randao_or_difficulty(self, f: Function) -> List[Node]:
        ret = set()
        for node in f.nodes:
            for var in node.variables_read:
                if is_dependent(
                    var, SolidityVariableComposed("block.number"), node
                ) or is_dependent(
                    var, SolidityVariableComposed("block.timestamp"), node
                ):
                    ret.add(node)
            for ir in node.irs:
                if (
                    isinstance(ir, SolidityCall)
                    and ir.function == SolidityFunction("number()")
                    or isinstance(ir, SolidityCall)
                    and ir.function == SolidityFunction("timestamp()")
                ):
                    ret.add(node)
        return list(ret)

    def _detect(self) -> List[Output]:
        """Main function"""
        results = []

        if os.getenv(ARBITRUM_KEY) is None:
            return results

        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions_and_modifiers:
                nodes = self._find_randao_or_difficulty(f)
                if nodes:
                    info = [
                        f,
                        " function uses block.number/block.timestamp inside the code of the Arbitrum contract\n"
                        "They behave different than on Ethereum, for details: (https://docs.arbitrum.io/for-devs/concepts/differences-between-arbitrum-ethereum/block-numbers-and-time)\n",
                        "Verify, that contract's logic does not break because of these differences\n"
                        "\t Usages:\n",
                    ]

                    nodes.sort(key=lambda x: x.node_id)

                    for node in nodes:
                        info += ["\t- ", node, "\n"]

                    results.append(self.generate_result(info))
        return results



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from typing import List
from slither.core.cfg.node import Node
from slither.core.declarations import Function, SolidityVariableComposed
from slither.analyses.data_dependency.data_dependency import is_dependent


class NftApproveWarning(AbstractDetector):
    """
    Sees if contract contains erc721.[safe]TransferFrom(from, ...) where from parameter is not related to msg.sender
    """

    ARGUMENT = 'pess-nft-approve-warning' # slither will launch the detector with slither.py --detect nft-approve-warning
    HELP = '[safe]TransferFrom(from,...) - from parameter is not related to msg.sender'
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://ventral.digital/posts/2022/8/18/sznsdaos-bountyboard-unauthorized-transferfrom-vulnerability'
    WIKI_TITLE = 'NFT Approve Warning'
    WIKI_DESCRIPTION = "In [safe]TransferFrom() from parameter must be related to msg.sender"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'from parameter must be related to msg.sender'

    _signatures=["transferFrom(address,address,uint256)", "safeTransferFrom(address,address,uint256,bytes)", "safeTransferFrom(address,address,uint256)"]

    def _detect_arbitrary_from(self, f: Function):
        all_high_level_calls = [
            f_called[1].solidity_signature
            for f_called in f.high_level_calls
            if isinstance(f_called[1], Function)
        ]
        all_library_calls = [f_called[1].solidity_signature for f_called in f.library_calls]

        all_calls = all_high_level_calls + all_library_calls

        if (any(map(lambda s: s in all_calls, self._signatures))):
            return self._arbitrary_from(f.nodes)
        else:
            return []

    def _arbitrary_from(self, nodes: List[Node]):
        """Finds instances of (safe)transferFrom that do not use msg.sender or address(this) as from parameter."""
        irList = []
        for node in nodes:
            for ir in node.irs:
                if (
                    hasattr(ir, 'function')
                    and hasattr(ir.function, 'solidity_signature')
                    and ir.function.solidity_signature in self._signatures
                ):
                    is_from_sender = is_dependent(ir.arguments[0], SolidityVariableComposed("msg.sender"), node.function.contract)
                    # is_from_self = is_dependent(ir.arguments[0], SolidityVariable("this"), node.function.contract)
                    if (not is_from_sender): # and not is_from_self
                        irList.append(ir.node)
        return irList


    def _detect(self):
        """Detect transfers that use arbitrary `from` parameter."""
        res = []
        for c in self.compilation_unit.contracts_derived:
            for f in c.functions:
                for d in self._detect_arbitrary_from(f):
                    res.append(self.generate_result([f.contract_declarer.name, ' ',f.name, ' parameter from is not related to msg.sender ', d, '\n']))
        return res



from typing import List, Optional
from slither.core.cfg.node import NodeType, Node
from slither.core.declarations import Contract, Function
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output


class Context():
    def __init__(self):
      self.visited = []
      self.depth = 0
      self.index_var = None
      self.father_type = None
      self.continue_seen = False
      self.checked_increase = False # necessary condition is the absence of a checked increase.
      self.unchecked_increase = False # Used to detect when the unchecked{i++} happens before the continue statement (false pos to filter out).
      #                                 We use this also in the case of [continue, break, return] at the end of the loop body, in such case the 
      #                                 presence of an unchecked increase is required.
      
    def __str__(self): # debug
        return 'index_var: ' + str(self.index_var) + ' - continue: ' + str(self.continue_seen) + ' - checked_increase: ' + str(self.checked_increase)


def check_contract(c: Contract) -> List[Function]:
  results_raw: List[Function] = []

  # skip test contracts
  if any(x in c.name for x in ["Test", "Mock"]): 
    return results_raw

  # filter unwanted funcs
  to_inspect = (x for x in c.functions if x.is_implemented)

  for f in to_inspect: 
    if check_function(f.entry_point, Context()):
      results_raw.append(f)
  return results_raw


def check_function(node: Optional[Node], ctx: Context)-> bool:
  if node is None: 
    return []

  if node in ctx.visited:
    return []
  ctx.visited.append(node)
  
  # continue_seen flag
  if node.type == NodeType.CONTINUE and ctx.index_var is not None and not ctx.unchecked_increase: # ctx.index_var is not None makes sure we are inside the loop
    ctx.continue_seen = True

  # set loop index variable
  if node.type == NodeType.STARTLOOP:
    if len(node.fathers) > 0: # should always be true
      main_father = node.fathers[0] 
      local_written = main_father.local_variables_written 
      if not len(local_written): # this happens if the index is left implicitly to 0 e.g. for (uint i; i < ...
        local_written = [main_father.variable_declaration] # in this case we get the declared one
      if len(local_written) == 1: # should always be true
        ctx.index_var = local_written[0] # strong assumption, it might be wrong in weird situations

  # check increase flag
  if node.type == NodeType.EXPRESSION and ctx.index_var is not None:
    local_written = node.local_variables_written
    if ctx.index_var in local_written: 
      if node.scope.is_checked:
        ctx.checked_increase = True
      else:
        ctx.unchecked_increase = True

  # if end_loop and not index increased, return true
  if node.type == NodeType.ENDLOOP:
    if ctx.index_var is not None and ctx.continue_seen and not ctx.checked_increase:
      if ctx.father_type in [NodeType.BREAK, NodeType.RETURN, NodeType.CONTINUE]:
        if ctx.unchecked_increase:
          return True
      else:
        return True 
    else:
      # clean the context for the next loop if any
      ctx.index_var = None 
      ctx.continue_seen = False 
      ctx.checked_increase = False

  # if didn't return true, inspect sons
  ctx.father_type = node.type
  ret = False
  for son in node.sons:
    ret = ret or check_function(son, ctx) # ctx passed by reference to collect 'continue_seen' and 'checked_increase' flags among all the previous branches
  return ret


class ForContinueIncrement(AbstractDetector):

    ARGUMENT = "pess-for-continue-increment" # slither will launch the detector with slither.py --detect for-continue-increment
    HELP = "Loop might turn out infinite"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.LOW

    WIKI = "https://github.com/sherlock-audit/2023-02-olympus-judging/issues/125"

    WIKI_TITLE = "For continue increment"
    WIKI_DESCRIPTION = "A continue statement before an unchecked index increment can turn into an infinite loop"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Increment the loop index before the continue statement"

    def _detect(self) -> List[Output]:
      """"""
      results: List[Output] = []
      results_raw: List[Function] = []
      for c in self.compilation_unit.contracts_derived:
        results_raw += check_contract(c)

      for sl in results_raw:
        info = [
          f"Found vulnerable for loop in function '",
          sl,
          "'\n"
        ]
        res = self.generate_result(info)
        results.append(res)

      return results         




from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function


class TokenFallback(AbstractDetector):
    """
    Sees if a token contract has a fallback function.
    """

    ARGUMENT = 'pess-token-fallback' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'The contract has a fallback function'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/token_fallback.md'
    WIKI_TITLE = 'Token Fallback'
    WIKI_DESCRIPTION = "Fallback function in tokens must be used carefully"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Make sure that contract can not be exploited with fallback function'


    def _has_fallback(self, fun: Function) -> bool: 
        """Checks if token contract has a fallback function"""
        if fun.name == "fallback":
            return True
        return False

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if contract.is_token:
                for function in contract.functions:
                    x = self._has_fallback(function)
                    if x:
                        res.append(self.generate_result([
                            "Token contract has a ",
                            function, ' which might be dangerous to implement',
                            '\n']))
        return res



from enum import Enum
from typing import List, Tuple
from collections import namedtuple

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    classification_txt,
)
from slither.slithir.operations import SolidityCall, HighLevelCall
from slither.core.declarations import (
    Contract,
    SolidityVariableComposed,
    FunctionContract,
)
from slither.core.cfg.node import Node
from slither.slithir.operations import LowLevelCall
from slither.analyses.data_dependency.data_dependency import is_dependent, is_tainted


# TODO/Possible improvements:
# look for transferFroms if there is any transferFrom and contract contains whole arbitrary call - it is screwed
# Filter out role protected

DetectorParams = namedtuple(
    "DetectorParams", ["argument_suffix", "impact", "confidence"]
)


class ArbitraryCallDetectors:
    ArbitraryCallWithApproveStored = DetectorParams(
        "-with-stored-erc20-approves",
        DetectorClassification.HIGH,
        DetectorClassification.HIGH,
    )
    ArbitraryCall = DetectorParams(
        "",
        DetectorClassification.HIGH,
        DetectorClassification.HIGH,
    )
    ArbitraryCallDestinationTainted = DetectorParams(
        "-destination-tainted",
        DetectorClassification.MEDIUM,
        DetectorClassification.MEDIUM,
    )
    ArbitraryCallCalldataTainted = DetectorParams(
        "-calldata-tainted",
        DetectorClassification.MEDIUM,
        DetectorClassification.MEDIUM,
    )


class ArbitraryCall(AbstractDetector):
    """
    Detects arbitrary and dangerous calls.
    (only detects low-level calls)
    """

    ARGUMENT = "pess-arbitrary-call"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "someaddress.call(somedata)"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = (
        "https://github.com/pessimistic-io/slitherin/blob/master/docs/arbitrary_call.md"
    )
    WIKI_TITLE = "Arbitrary calls"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = "Attacker can manipulate on inputs"
    WIKI_RECOMMENDATION = "Do not allow arbitrary calls"

    def _is_role_protected(self, function: FunctionContract):
        for m in function.modifiers:
            if m.name.startswith("only"):
                return True
        return False

    def analyze_function(
        self, function: FunctionContract
    ) -> Tuple[
        List[Tuple[FunctionContract, Node, LowLevelCall, bool, bool]], bool
    ]:  # TODO(yhtiyar): make return type as named tuple/class
        results = []
        stores_approve = False
        for node in function.nodes:
            for ir in node.irs:
                try:
                    destination_tainted = False
                    args_tainted = False
                    if isinstance(ir, LowLevelCall):
                        destination_tainted = is_tainted(ir.destination, node, True)
                        if (
                            destination_tainted
                            and ir.destination == SolidityVariableComposed("msg.sender")
                        ):
                            # We don't care about msg.sender, since will be only reentrancy issue
                            destination_tainted = False
                        args_tainted = any(
                            is_tainted(arg, node, True) for arg in ir.arguments
                        )  # seems like ir.arguments = [data] for all low-level calls

                    elif (
                        isinstance(ir, SolidityCall)
                        and ir.function.name
                        == "delegatecall(uint256,uint256,uint256,uint256,uint256,uint256)"
                    ):
                        # delegatecall
                        destination_tainted = is_tainted(ir.arguments[1], node, True)
                        # #args_tainted = is_tainted(ir.arguments[2], node, True)
                        # for delegateCall we don't actually care about args, since
                        # for all proxies, user fully sets args
                    elif (
                        isinstance(ir, SolidityCall)
                        and ir.function.name
                        == "call(uint256,uint256,uint256,uint256,uint256,uint256,uint256)"
                    ):
                        # call()
                        destination_tainted = is_tainted(ir.arguments[1], node, True)
                        args_tainted = is_tainted(ir.arguments[3], node, True)

                    elif (
                        isinstance(ir, HighLevelCall)
                        and ir.function.name == "transferFrom"
                    ):
                        stores_approve = True

                    if args_tainted or destination_tainted:
                        results.append(
                            (
                                function,
                                node,
                                ir,
                                args_tainted,
                                destination_tainted,
                            )
                        )
                except Exception as e:
                    print("ArbitraryCall:Failed to check types", e)
                    print(ir)

        return (results, stores_approve)

    def analyze_contract(self, contract: Contract):
        stores_approve = False
        all_tainted_calls: List[
            Tuple[FunctionContract, Node, LowLevelCall, bool, bool]
        ] = []
        results = []
        for f in contract.functions:
            (res, _stores_approve) = self.analyze_function(f)
            stores_approve |= _stores_approve
            if res:
                all_tainted_calls.extend(res)

        for call_fn, node, ir, args_tainted, destination_tainted in all_tainted_calls:
            info = ["Manipulated call found: ", node, " in ", call_fn, "\n"]
            if args_tainted and destination_tainted:
                text = "Both destination and calldata could be manipulated"
            else:
                part = "calldata" if args_tainted else "destination"
                text = f"Only the {part} could be manipulated"
            info += [f"{text}\n"]

            for f in contract.functions:
                if f.visibility not in ["external", "public"]:
                    continue
                if self._is_role_protected(f):
                    continue

                fn_taints_args = False
                fn_taints_destination = False
                args = (
                    ir.arguments[0] if isinstance(ir, LowLevelCall) else ir.arguments[3]
                )
                if args_tainted and any(
                    is_dependent(args, fn_arg, node) for fn_arg in f.variables
                ):
                    fn_taints_args = True

                destination = (
                    ir.destination if isinstance(ir, LowLevelCall) else ir.arguments[1]
                )
                if destination_tainted and any(
                    is_dependent(destination, fn_arg, node) for fn_arg in f.variables
                ):
                    fn_taints_destination = True

                if not (fn_taints_args or fn_taints_destination):
                    continue

                detectorParams: DetectorParams = None
                if fn_taints_args and fn_taints_destination:
                    if stores_approve:
                        text = "The call could be fully manipulated (arbitrary call). This contract also STORES APPROVES!!!"
                        detectorParams = (
                            ArbitraryCallDetectors.ArbitraryCallWithApproveStored
                        )

                    else:
                        text = "The call could be fully manipulated (arbitrary call)"
                        detectorParams = ArbitraryCallDetectors.ArbitraryCall
                else:
                    if fn_taints_args:
                        part = "calldata"
                        detectorParams = (
                            ArbitraryCallDetectors.ArbitraryCallCalldataTainted
                        )

                    else:
                        part = "destination"
                        detectorParams = (
                            ArbitraryCallDetectors.ArbitraryCallDestinationTainted
                        )

                    text = f"The {part} could be manipulated"
                info += [f"\t{text} through ", f, "\n"]

                res = self.generate_result(info)
                res.add(node)
                res.data["check"] = self.ARGUMENT + detectorParams.argument_suffix
                res.data["impact"] = classification_txt[detectorParams.impact]
                res.data["confidence"] = classification_txt[detectorParams.confidence]

                results.append(res)
        return results

    def _detect(self):
        results = []
        for contract in self.compilation_unit.contracts_derived:
            r = self.analyze_contract(contract)
            if r:
                results.extend(r)
        return results



from slither.core.cfg.node import NodeType
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import SolidityVariableComposed
from slither.analyses.data_dependency.data_dependency import is_dependent


class OnlyEOACheck(AbstractDetector):
    """
    Shows expression msg.sender == tx.origin
    """

    ARGUMENT = 'pess-only-eoa-check' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'msg.sender == tx.origin'
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/falsy_only_eoa_modifier.md'
    WIKI_TITLE = 'Falsy Only EOA Modifier'
    WIKI_DESCRIPTION = "Logic msg.sender == tx.origin must be removed"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Do not rely on logic msg.sender == tx.origin when protecting contract logic'

    def hasWrongEq(self, fun, params=None):
        varListMsg = []
        varListTx = []
        for n in fun.nodes: # в первом приближении нода это строчка
            if(n.type==NodeType.IF):
                for var in n.solidity_variables_read:
                    is_msg = is_dependent(var, SolidityVariableComposed("msg.sender"), n.function.contract)
                    if is_msg:
                        varListMsg.append(var)
                    is_tx = is_dependent(var, SolidityVariableComposed("tx.origin"), n.function.contract)
                    if is_tx: 
                        varListTx.append(var)
                for i in range(len(varListTx)):
                    for j in range(len(varListMsg)):
                        if(str(n).__contains__(f'{varListMsg[j]} == {varListTx[i]}') or str(n).__contains__(f'{varListTx[i]} == {varListMsg[j]}')):
                            return "True"
        return "False"

    def _detect(self):
        res = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                x = self.hasWrongEq(f)
                if (x != "False"):
                    res.append(self.generate_result([
                        f.contract_declarer.name, ' ',
                        f.name, ' has a falsy EOA modifier ',
                        x, ' is set'
                        '\n']))
        return res



import re
from typing import List, Tuple

from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import SolidityCall, Condition
from slither.core.declarations import (
    FunctionContract,
    SolidityVariableComposed,
    Function,
)
from slither.core.cfg.node import Node
from slither.slithir.operations import LowLevelCall
from slither.analyses.data_dependency.data_dependency import is_dependent


class AAVEFlashloanCallbackDetector(AbstractDetector):
    """
    Detects if the flashloan callback `executeOperation` has `initiator` and `msg.sender` validation
    """

    ARGUMENT = "pess-aave-flashloan-callback"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "see `executeOperation`callback docs"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/aave/flashloan_callback.md"
    WIKI_TITLE = "AAVE Flashloan callback"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = (
        "Attacker can directly call or initiate flashloan to this address"
    )
    WIKI_RECOMMENDATION = "Validate the msg.sender and `initiator` argument"

    def _analyze_function(
        self,
        function: FunctionContract,
        args_to_verify,
        check_recursive=False,
        visited=[],
    ) -> Tuple[set, set]:
        checked = set()
        unchecked = set(args_to_verify)

        for node in function.nodes:
            if node.is_conditional(include_loop=False):  # contains if/assert/require
                for a in unchecked:
                    if any([is_dependent(arg, a, node) for arg in node.variables_read]):
                        checked.add(a)

        unchecked.difference_update(checked)  # remove all checked

        visited.append(function)
        if check_recursive:
            for node in function.nodes:
                for internal_call in node.internal_calls:
                    # Filter to Function, as internal_call can be a solidity call
                    if isinstance(internal_call, Function):
                        _checked, _unchecked = self._analyze_function(
                            internal_call, unchecked, False, visited
                        )  # We are going only 1 step into in recursion
                        checked |= _checked
                        unchecked = _unchecked

        return checked, unchecked

    def _detect(self):
        results = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                if (
                    f.signature_str
                    == "executeOperation(address[],uint256[],uint256[],address,bytes) returns(bool)"
                ):
                    _, unchecked = self._analyze_function(
                        f,
                        [
                            SolidityVariableComposed("msg.sender"),
                            f.parameters[3],  # initiator parameter
                        ],
                        check_recursive=True,
                    )
                    if unchecked:
                        info = [
                            "Unchecked function parameters in AAVE callback: ",
                            f,
                            "\n",
                        ]
                        for var in unchecked:
                            if var == SolidityVariableComposed("msg.sender"):
                                info += ["\t", "'msg.sender' is not checked", "\n"]
                            else:
                                info += ["\t'", var.name, "' is not checked\n"]

                        tres = self.generate_result(info)
                        # for var in unchecked:
                        #     tres.add(var)
                        # tres.add(f)
                        results.append(tres)
        return results



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function


class TxGaspriceWarning(AbstractDetector):
    """
    Sees if contract function uses variable tx.gasprice.
    """

    ARGUMENT = 'pess-tx-gasprice' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'The contract uses tx.gasprice variable'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/tx_gasprice_warning.md'
    WIKI_TITLE = 'TX Gasprice Warning'
    WIKI_DESCRIPTION = "Tx.gasprice variable must be used carefully"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Make sure that tx.gasprice varible which is set by users does not exploit the contract logic'


    def _has_tx_gasprice(self, fun: Function) -> bool: 
        """Checks if function has tx.gasprice variable"""
        if isinstance(fun, Function):   # check for a correct function type
            for n in fun.nodes: # checks every node for a tx.gasprice variable
                if "tx.gasprice" in str(n):
                    return True
        return False


    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions_and_modifiers_declared:
                x = self._has_tx_gasprice(f)
                if x:
                    res.append(self.generate_result([
                        "In a function ",
                        f, ' tx.gasprice (which is set by user) variable is used. Make sure that it can not exploit the contract logic.',
                        '\n']))
        return res



import argparse
import logging
from typing import List, Any, Optional
import subprocess
import os
import shutil
import pty
import sys
from pathlib import Path
import slitherin
from pkg_resources import iter_entry_points

from .consts import *


def slitherin_detectors_list_as_arguments() -> str:
    return ",".join([detector.ARGUMENT for detector in slitherin.plugin_detectors])


def arbitrum_detectors_list_as_arguments() -> str:
    return ",".join([detector.ARGUMENT for detector in slitherin.artbitrum_detectors])


logging.basicConfig()
LOGGER = logging.getLogger("slitherinLogger")
LOGGER.setLevel(logging.INFO)

output_bytes = []


def read(fd):
    data = os.read(fd, 1024)
    output_bytes.append(data)
    return data


# this is modified version from : https://github.com/crytic/crytic-compile/blob/master/crytic_compile/utils/subprocess.py#L14
def run(
    cmd: List[str],
    **kwargs: Any,
) -> Optional[subprocess.CompletedProcess]:
    subprocess_cwd = Path(os.getcwd()).resolve()
    subprocess_exe = shutil.which(cmd[0])

    if subprocess_exe is None:
        LOGGER.error("Cannot execute `%s`, is it installed and in PATH?", cmd[0])
        return None

    LOGGER.info(
        "'%s' running (wd: %s)",
        " ".join(cmd),
        subprocess_cwd,
    )

    master, slave = pty.openpty()

    try:
        process = subprocess.Popen(cmd, stdout=slave, stderr=slave, text=False)
        os.close(slave)

        while True:
            try:
                output = os.read(master, 1024)
            except OSError as e:
                if e.errno == 5:  # Errno 5 corresponds to Input/output error
                    break
                else:
                    raise
            if not output:
                break
            decoded_output = output.decode(sys.stdout.encoding, errors="replace")
            print(decoded_output, end="")

        return_code = process.wait()
        if return_code != 0:
            raise Exception(
                f"Errored out with code: {return_code}, while running slither"
            )

    except Exception as e:
        print(f"Failed to run slither: {str(e)}")
        raise e


def handle_list() -> None:
    detectors = slitherin.plugin_detectors
    for detector in detectors:
        print(detector.ARGUMENT)

    print("\nTo pass as argument (for --detect/--exclude):")
    print(slitherin_detectors_list_as_arguments())


def handle_parser(args: argparse.Namespace, slither_args) -> None:
    slitherin_detectors = slitherin_detectors_list_as_arguments()
    slither_with_args = [
        "slither",
        "--fail-none",
    ] + slither_args  # using fail-none flag, so slither will return 0 even though there are findings

    if args.pess:
        run(
            slither_with_args + ["--detect", slitherin_detectors],
        )
    elif args.slither:
        run(
            slither_with_args + ["--exclude", slitherin_detectors],
        )
    elif args.separated:
        print("Only slither results:")
        run(
            slither_with_args + ["--exclude", slitherin_detectors],
        )
        print("Only slitherin results:")
        run(
            slither_with_args + ["--ignore-compile", "--detect", slitherin_detectors],
        )
    elif args.arbitrum:
        os.environ["SLITHERIN_ARBITRUM"] = "True"
        run(slither_with_args + ["--detect", arbitrum_detectors_list_as_arguments()])
        del os.environ["SLITHERIN_ARBITRUM"]

    else:
        run(slither_with_args)


def generate_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="""
        slitherin: Additional slither detectors by pessimistic.io
        All additional parameters will be passed to slither
        Usage examples:
            slitherin --pess PATH
            slitherin --pess PATH --json result.json
            slitherin --pess PATH --ignore-compile 
        """,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"slitherin {SLITHERIN_VERSION}",
        help="Show version",
    )

    parser.add_argument(
        "--pess",
        action="store_true",
        help="Run only pessimistic.io (slitherin) detectors",
    )

    parser.add_argument(
        "--slither",
        action="store_true",
        help="Run only slither detectors",
    )

    parser.add_argument(
        "--separated",
        action="store_true",
        help="Run slither detectors, then slitherin",
    )

    parser.add_argument(
        "--arbitrum",
        action="store_true",
        help="Run arbitrum detectors",
    )
    return parser


def main() -> None:
    """
    Handler for the "slitherin" command.
    """

    parser = generate_argument_parser()
    parsed, unknown = parser.parse_known_args()

    # It turned out that argparse has no solution to parse unkown args and subcommands,
    # so u need to parse subcommands manually
    if unknown and unknown[0] == "list":
        handle_list()
        return
    if not unknown and not parsed._get_args():
        parser.print_help()
        return
    handle_parser(parsed, unknown)


if __name__ == "__main__":
    main()



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification


class DoubleEntryTokenPossiblity(AbstractDetector):
    """
    Sees if contract contains a function which is vulnurable to double-entry tokens attack
    """

    ARGUMENT = 'pess-double-entry-token-alert' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'The function might be sensitive to double entry token usage'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/double_entry_token_possibility.md'
    WIKI_TITLE = 'Double-entry token'
    WIKI_DESCRIPTION = "Double-entry token related attack might be possible"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Project contract must be sustainable to a token which has two pointing addresses'


    def get_tokens_as_params(self, fun):

        res = []  # параметры функции
        for p in fun.parameters:
            if len(p._name) and str(p.type) in ['IERC20[]', 'address[]']:
                res.append(p)

        return res

    def do_have_token_interaction(self, fun, token):

        for n in fun.nodes:
            if str(token) in str(n.expression):
                if '.transfer' in str(n.expression): return True
                if '.balanceOf' in str(n.expression): return True

        return False

    def _detect(self):

        res = []

        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:

                tokens = self.get_tokens_as_params(f)

                if len(tokens)>0:
                    for t in tokens:
                        if self.do_have_token_interaction(f, t):
                            res.append(self.generate_result([
                                f.contract_declarer.name, ' ',
                                f, ' might be vulnerable to double-entry token exploit',
                                '\n']))

        return res


""""
    Re-entrancy detection
    Based on heuristics, it may lead to FP and FN
    Iterate over all the nodes of the graph until reaching a fixpoint
"""
from collections import namedtuple, defaultdict
from typing import Dict, List, Set
from slither.core.variables.variable import Variable
from slither.core.declarations import Function
from slither.core.cfg.node import NodeType, Node, Contract
from slither.detectors.abstract_detector import DetectorClassification
from .reentrancy.reentrancy import (
    Reentrancy,
    to_hashable,
    AbstractState,
    union_dict,
    _filter_if,
    is_subset,
)
from slither.slithir.operations import EventCall

FindingKey = namedtuple("FindingKey", ["function", "calls"])
FindingValue = namedtuple("FindingValue", ["variable", "written_at", "node", "nodes"])


def are_same_contract(a: Contract, b: Contract) -> bool:
    """
    Checks if A==B or A inherits from B or otherwise
    """
    return a == b or (b in a.inheritance) or (b in a.derived_contracts)


class ReadOnlyReentrancyState(AbstractState):
    def __init__(self):
        super().__init__()
        self._reads_external: Dict[Variable, Set[Node]] = defaultdict(set)
        self._reads_external_contract_list: Dict[Variable, Set[Contract]] = defaultdict(
            set
        )
        self._written_external: Dict[Variable, Set[Node]] = defaultdict(set)
        self._written: Dict[Variable, Set[Node]] = defaultdict(set)

    @property
    def reads_external(self) -> Dict[Variable, Set[Node]]:
        return self._reads_external

    @property
    def reads_external_contract_list(self) -> Dict[Variable, Set[Contract]]:
        return self._reads_external_contract_list

    @property
    def written_external(self) -> Dict[Variable, Set[Node]]:
        return self._written_external

    @property
    def written(self) -> Dict[Variable, Set[Node]]:
        return self._written

    def add(self, fathers):
        super().add(fathers)
        self._reads_external = union_dict(self._reads_external, fathers.reads_external)
        self._reads_external_contract_list = union_dict(
            self._reads_external_contract_list, fathers.reads_external_contract_list
        )

    def does_not_bring_new_info(self, new_info):
        return (
            super().does_not_bring_new_info(new_info)
            and is_subset(new_info.reads_external, self._reads_external)
            and is_subset(
                new_info.reads_external_contract_list,
                self._reads_external_contract_list,
            )
        )

    def merge_fathers(self, node, skip_father, detector):
        for father in node.fathers:
            if detector.KEY in father.context:
                self._send_eth = union_dict(
                    self._send_eth,
                    {
                        key: values
                        for key, values in father.context[detector.KEY].send_eth.items()
                        if key != skip_father
                    },
                )
                self._calls = union_dict(
                    self._calls,
                    {
                        key: values
                        for key, values in father.context[detector.KEY].calls.items()
                        if key != skip_father
                    },
                )
                self._reads = union_dict(
                    self._reads, father.context[detector.KEY].reads
                )
                self._reads_external = union_dict(
                    self._reads_external, father.context[detector.KEY].reads
                )
                self._written_external = union_dict(
                    self._written_external, father.context[detector.KEY].reads
                )

    def analyze_node(self, node: Node, detector):
        state_vars_read: Dict[Variable, Set[Node]] = defaultdict(
            set, {v: {node} for v in node.state_variables_read}
        )

        # All the state variables written
        state_vars_written: Dict[Variable, Set[Node]] = defaultdict(
            set, {v: {node} for v in node.state_variables_written}
        )

        external_state_vars_read: Dict[Variable, Set[Node]] = defaultdict(set)
        external_state_vars_written: Dict[Variable, Set[Node]] = defaultdict(set)
        external_state_vars_read_contract_list: Dict[
            Variable, Set[Contract]
        ] = defaultdict(set)

        slithir_operations = []
        # Add the state variables written in internal calls
        for internal_call in node.internal_calls:
            # Filter to Function, as internal_call can be a solidity call
            if isinstance(internal_call, Function):
                for internal_node in internal_call.all_nodes():
                    for read in internal_node.state_variables_read:
                        state_vars_read[read].add(internal_node)
                    for write in internal_node.state_variables_written:
                        state_vars_written[write].add(internal_node)
                slithir_operations += internal_call.all_slithir_operations()

        for contract, v in node.high_level_calls:
            if isinstance(v, Function):
                for internal_node in v.all_nodes():
                    for read in internal_node.state_variables_read:
                        external_state_vars_read[read].add(internal_node)
                        external_state_vars_read_contract_list[read].add(contract)

                    if internal_node.context.get(detector.KEY):
                        for r in internal_node.context[detector.KEY].reads_external:
                            external_state_vars_read[r].add(internal_node)
                            external_state_vars_read_contract_list[r].add(contract)
                    for write in internal_node.state_variables_written:
                        external_state_vars_written[write].add(internal_node)

        contains_call = False

        self._written = state_vars_written
        self._written_external = external_state_vars_written
        for ir in node.irs + slithir_operations:
            if detector.can_callback(ir):
                self._calls[node] |= {ir.node}
                contains_call = True

            if detector.can_send_eth(ir):
                self._send_eth[node] |= {ir.node}

            if isinstance(ir, EventCall):
                self._events[ir] |= {ir.node, node}

        self._reads = union_dict(self._reads, state_vars_read)
        self._reads_external = union_dict(
            self._reads_external, external_state_vars_read
        )
        self._reads_external_contract_list = union_dict(
            self._reads_external_contract_list, external_state_vars_read_contract_list
        )

        return contains_call


class ReadOnlyReentrancy(Reentrancy):
    ARGUMENT = "pess-readonly-reentrancy"
    HELP = "Read-only reentrancy vulnerabilities"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/readonly_reentrancy.md"
    WIKI_TITLE = "Read-only reentrancy vulnerabilities"
    WIKI_DESCRIPTION = "Check docs"
    STANDARD_JSON = False
    KEY = "readonly_reentrancy"

    contracts_read_variable: Dict[Variable, Set[Contract]] = defaultdict(set)
    contracts_written_variable_after_reentrancy: Dict[
        Variable, Set[Contract]
    ] = defaultdict(set)

    def _explore(self, node, visited, skip_father=None):
        if node in visited:
            return

        visited = visited + [node]

        fathers_context = ReadOnlyReentrancyState()
        fathers_context.merge_fathers(node, skip_father, self)

        # Exclude path that dont bring further information
        if node in self.visited_all_paths:
            if self.visited_all_paths[node].does_not_bring_new_info(fathers_context):
                return
        else:
            self.visited_all_paths[node] = ReadOnlyReentrancyState()

        self.visited_all_paths[node].add(fathers_context)

        node.context[self.KEY] = fathers_context

        contains_call = fathers_context.analyze_node(node, self)
        node.context[self.KEY] = fathers_context

        sons = node.sons
        if contains_call and node.type in [NodeType.IF, NodeType.IFLOOP]:
            if _filter_if(node):
                son = sons[0]
                self._explore(son, visited, node)
                sons = sons[1:]
            else:
                son = sons[1]
                self._explore(son, visited, node)
                sons = [sons[0]]

        for son in sons:
            self._explore(son, visited)

    def find_writes_after_reentrancy(self):
        written_after_reentrancy: Dict[Variable, Set[Node]] = defaultdict(set)
        written_after_reentrancy_external: Dict[Variable, Set[Node]] = defaultdict(set)
        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                for node in f.nodes:
                    # dead code
                    if self.KEY not in node.context:
                        continue
                    if node.context[self.KEY].calls:
                        if not any(n != node for n in node.context[self.KEY].calls):
                            continue
                        # TODO: check if written items exist
                        for v, nodes in node.context[self.KEY].written.items():
                            written_after_reentrancy[v].add(node)
                            self.contracts_written_variable_after_reentrancy[v].add(
                                contract
                            )
                        for v, nodes in node.context[self.KEY].written_external.items():
                            written_after_reentrancy_external[v].add(node)
                            self.contracts_written_variable_after_reentrancy[v].add(
                                contract
                            )

        return written_after_reentrancy, written_after_reentrancy_external

    # IMPORTANT:
    # FOR the external reads, that var should be external written in the same contract
    def get_readonly_reentrancies(self):
        (
            written_after_reentrancy,
            written_after_reentrancy_external,
        ) = self.find_writes_after_reentrancy()
        result = defaultdict(set)

        warnings = defaultdict(set)

        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                for node in f.nodes:

                    if self.KEY not in node.context:
                        continue
                    vulnerable_variables = set()
                    warning_variables = set()
                    for r, nodes in node.context[self.KEY].reads.items():

                        if r in written_after_reentrancy:
                            finding_value = FindingValue(
                                r,
                                tuple(
                                    sorted(
                                        list(written_after_reentrancy[r]),
                                        key=lambda x: x.node_id,
                                    )
                                ),
                                node,
                                tuple(sorted(nodes, key=lambda x: x.node_id)),
                            )
                            if are_same_contract(r.contract, f.contract):
                                if f.view and f.visibility in ["public", "external"]:
                                    warning_variables.add(finding_value)
                            else:
                                vulnerable_variables.add(finding_value)

                    for r, nodes in node.context[self.KEY].reads_external.items():
                        if are_same_contract(r.contract, f.contract):
                            # TODO(yhtiyar): In case f.view we can notify the user that the given
                            # method could be vulnerable if other contract will use it
                            continue
                        if r in written_after_reentrancy_external:
                            isVulnerable = any(
                                c in self.contracts_written_variable_after_reentrancy[r]
                                for c in node.context[
                                    self.KEY
                                ].reads_external_contract_list[r]
                            )
                            if isVulnerable:
                                vulnerable_variables.add(
                                    FindingValue(
                                        r,
                                        tuple(
                                            sorted(
                                                list(
                                                    written_after_reentrancy_external[r]
                                                ),
                                                key=lambda x: x.node_id,
                                            )
                                        ),
                                        node,
                                        tuple(sorted(nodes, key=lambda x: x.node_id)),
                                    )
                                )

                        if r in written_after_reentrancy:
                            vulnerable_variables.add(
                                FindingValue(
                                    r,
                                    tuple(
                                        sorted(
                                            list(written_after_reentrancy[r]),
                                            key=lambda x: x.node_id,
                                        )
                                    ),
                                    node,
                                    tuple(sorted(nodes, key=lambda x: x.node_id)),
                                )
                            )

                    if vulnerable_variables:
                        finding_key = FindingKey(
                            function=f, calls=to_hashable(node.context[self.KEY].calls)
                        )
                        result[finding_key] |= vulnerable_variables
                    if warning_variables:
                        finding_key = FindingKey(
                            function=f, calls=to_hashable(node.context[self.KEY].calls)
                        )
                        warnings[finding_key] |= warning_variables
        return result, warnings

    def _gen_results(self, raw_results, info_text):
        results = []

        result_sorted = sorted(
            list(raw_results.items()), key=lambda x: x[0].function.name
        )

        varsRead: List[FindingValue]
        for (func, calls), varsRead in result_sorted:

            varsRead = sorted(varsRead, key=lambda x: (x.variable.name, x.node.node_id))

            info = [f"{info_text} ", func, ":\n"]

            info += [
                "\tState variables read that were written after the external call(s):\n"
            ]
            for finding_value in varsRead:
                info += [
                    "\t- ",
                    finding_value.variable,
                    " was read at ",
                    finding_value.node,
                    "\n",
                ]
                # info += ["\t- ", finding_value.node, "\n"]

                # for other_node in finding_value.nodes:
                #     if other_node != finding_value.node:
                #         info += ["\t\t- ", other_node, "\n"]

                # TODO: currently we are not printing the whole call-stack of variable
                # it wasn't working properly, so I am removing it for now to avoid confusion

                info += ["\t\t This variable was written at (after external call):\n"]
                for other_node in finding_value.written_at:
                    # info += ["\t- ", call_info, "\n"]
                    if other_node != finding_value.node:
                        info += ["\t\t\t- ", other_node, "\n"]

            # Create our JSON result
            res = self.generate_result(info)

            res.add(func)

            # Add all variables written via nodes which write them.
            for finding_value in varsRead:
                res.add(
                    finding_value.node,
                    {
                        "underlying_type": "variables_written",
                        "variable_name": finding_value.variable.name,
                    },
                )
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        res.add(
                            other_node,
                            {
                                "underlying_type": "variables_written",
                                "variable_name": finding_value.variable.name,
                            },
                        )

            # Append our result
            results.append(res)

        return results

    def _detect(self):  # pylint: disable=too-many-branches
        results = []
        try:
            super()._detect()
            reentrancies, warnings = self.get_readonly_reentrancies()
            results += self._gen_results(reentrancies, "Readonly-reentrancy in ")
            results += self._gen_results(
                warnings,
                "Potential vulnerable to readonly-reentrancy function (if read in other function)",
            )
        except Exception as e:
            info = [
                "Error during detection of readonly-reentrancy:\n",
                "Please inform this to Yhtyyar\n",
                f"error details:",
                e,
            ]
        return results



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function


class MultipleStorageRead(AbstractDetector):
    """
    Sees if contract function reads the same storage variable multiple times.
    """

    ARGUMENT = 'pess-multiple-storage-read' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'The same contract storage variable is read multiple times in a function'
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/multiple_storage_read.md'
    WIKI_TITLE = 'Multiple Storage Read'
    WIKI_DESCRIPTION = "The same storage variable must not be read multiple times in the same function"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Consider creating and reading a local variable instead of a storage one if you read it more than once'


    def _has_multiple_storage_read(self, fun: Function) -> bool: 
        """Checks if function has multiple storage read"""
        if isinstance(fun, Function):   # check for a correct function type
            multiple_reads = []
            state_variables_used = []
            for n in fun.nodes: # check reads of state variables in nodes
                for node_state_variable in n.state_variables_read:
                    state_variables_used.append(node_state_variable)
            for function_state_variable in fun.state_variables_read: # check the number of times state variable are read
                indices = [i for i, x in enumerate(state_variables_used) if x == function_state_variable]
                if len(indices) > 2:
                    multiple_reads.append(function_state_variable)
            if multiple_reads:  
                return multiple_reads


    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions_and_modifiers_declared:
                x = self._has_multiple_storage_read(f)
                if x:
                    for multiple_read_var in x:
                        res.append(self.generate_result([
                            "In a function", ' ',
                            f, ' variable ', multiple_read_var , ' is read multiple times'
                            '\n']))
        return res



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.core.variables.variable import Variable


class UnprotectedInitialize(AbstractDetector):
    """
    Sees if a contract has initialize functions that are not protected by anything.
    """

    ARGUMENT = "pess-unprotected-initialize"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "The contract must have its initialize function protected"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/unprotected_initialize.md"
    WIKI_TITLE = "Unprotected Initialize"
    WIKI_DESCRIPTION = "Initializers must be protected"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Protect initializers with modifiers/require statements"

    def _is_initialize(self, fun: Function) -> bool:
        """Checks if function is initialize one"""
        if isinstance(fun, Function) and "init" in fun.name:
            return True
        return False

    def _has_modifiers(self, fun: Function) -> bool:
        """Checks if function has modifier protection"""
        for modifier in fun.modifiers:
            if str(modifier) == "onlyOwner" or str(modifier) == "initializer":
                return True
        return False

    def _has_require(self, fun: Function) -> bool:
        """Checks if function has require statement protection"""
        for node in fun.nodes:
            if "require" in str(node):
                for variable in node.variables_read:
                    if isinstance(variable, Variable):
                        if str(variable.type) == "address":
                            return True
        return False
    
    def _has_if_with_reverts(self, fun: Function) -> bool:
        for node in fun.nodes:
            if node.contains_if():
                for child in node.sons:
                    if "revert()" in str(child) or "revert(string)" in str(child):
                        return True
        return False

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if contract.is_interface:
                continue
            for f in contract.functions_and_modifiers_declared:
                x = self._is_initialize(f)
                if x:
                    is_safe = self._has_modifiers(f)
                    is_safe2 = self._has_require(f)
                    is_safe3 = self._has_if_with_reverts(f)
                    if not is_safe and not is_safe2 and not is_safe3:
                        res.append(
                            self.generate_result(
                                [
                                    "Function ",
                                    f,
                                    " is an unprotected initializer.",
                                    "\n",
                                ]
                            )
                        )
        return res



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
import slither.core.expressions.new_array as na


class StrangeSetter(AbstractDetector):
    """
    Sees if contract contains a setter, that does not change contract storage variables or that does not use arguments for an external call.
    """

    ARGUMENT = "pess-strange-setter"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Contract storage parameter is not changed by setter"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/pessimistic-io/slitherin/blob/master/docs/strange_setter.md"
    )
    WIKI_TITLE = "Strange Setter"
    WIKI_DESCRIPTION = "Setter must write to storage variables or pass arguments to external calls"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Make sure that your setter actually sets something"

    def _is_strange_setter(self, fun: Function) -> bool:
        """Checks if setter sets smth to a storage variable and if function parameters are used when setting"""
        if not isinstance(fun, Function):
            return True

        if not fun.parameters:
            # nothing is in the params, so we don't care
            return False
        used_params = set()
        for (
            fin
        ) in fun.internal_calls:  # branch with for-loop for setters in internal calls
            if isinstance(fin, Function):
                for param in fin.parameters:
                    for n in fin.nodes:
                        if n.state_variables_written and str(param) in str(
                            n
                        ):  # check if there's a state variable setter using function parameters
                            used_params.add(param)
        for param in fun.parameters:
            if fun.state_variables_written:
                for n in fun.nodes:
                    if str(param) in str(n):
                        used_params.add(param)
        for param in fun.parameters:
            for external in fun.external_calls_as_expressions:
                if isinstance(external._called, na.NewArray):
                    continue
                for arg in [*external.arguments, external._called._expression]:
                    if str(arg) == str(param):
                        used_params.add(param)
        intersection_len = len(set(fun.parameters) & used_params)
        return intersection_len != len(fun.parameters)

    def _is_strange_constructor(self, fun: Function) -> bool:
        """Checks if constructor sets nothing"""
        return self._is_strange_setter(fun)

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            if not contract.is_interface:
                overriden_funtions = []  # functions that are overridden
                for f in contract.functions:
                    if f.parameters:
                        x = False
                        overriden_funtions.append(
                            contract.get_functions_overridden_by(f)
                        )  # adding functions to an overridden list
                        if f.name == "constructor":
                            x = self._is_strange_constructor(f)
                        elif (
                            f.name.startswith("set")
                            and not f in overriden_funtions
                            and len(f.nodes) != 0
                        ):  # check if setter starts with 'set', is not overriden and is not empty
                            x = self._is_strange_setter(f)
                        if x:
                            res.append(
                                self.generate_result(
                                    [
                                        "Function",
                                        " ",
                                        f,
                                        " is a strange setter. ",
                                        "Nothing is set in constructor or set in a function without using function parameters"
                                        "\n",
                                    ]
                                )
                            )
        return res



""""
    This is a copy paste from the original slither reentrancy v0.9.1



    Re-entrancy detection
    Based on heuristics, it may lead to FP and FN
    Iterate over all the nodes of the graph until reaching a fixpoint
"""
from collections import defaultdict
from typing import Set, Dict, Union

from slither.core.cfg.node import NodeType, Node
from slither.core.declarations import Function
from slither.core.expressions import UnaryOperation, UnaryOperationType
from slither.core.variables.variable import Variable
from slither.detectors.abstract_detector import AbstractDetector
from slither.slithir.operations import Call, EventCall


def union_dict(d1, d2):
    d3 = {
        k: d1.get(k, set()) | d2.get(k, set())
        for k in set(list(d1.keys()) + list(d2.keys()))
    }
    return defaultdict(set, d3)


def dict_are_equal(d1, d2):
    if set(list(d1.keys())) != set(list(d2.keys())):
        return False
    return all(set(d1[k]) == set(d2[k]) for k in d1.keys())


def is_subset(
    new_info: Dict[Union[Variable, Node], Set[Node]],
    old_info: Dict[Union[Variable, Node], Set[Node]],
):
    for k in new_info.keys():
        if k not in old_info:
            return False
        if not new_info[k].issubset(old_info[k]):
            return False
    return True


def to_hashable(d: Dict[Node, Set[Node]]):
    list_tuple = list(
        tuple((k, tuple(sorted(values, key=lambda x: x.node_id))))
        for k, values in d.items()
    )
    return tuple(sorted(list_tuple, key=lambda x: x[0].node_id))


class AbstractState:
    def __init__(self):
        # send_eth returns the list of calls sending value
        # calls returns the list of calls that can callback
        # read returns the variable read
        # read_prior_calls returns the variable read prior a call
        self._send_eth: Dict[Node, Set[Node]] = defaultdict(set)
        self._calls: Dict[Node, Set[Node]] = defaultdict(set)
        self._reads: Dict[Variable, Set[Node]] = defaultdict(set)
        self._reads_prior_calls: Dict[Node, Set[Variable]] = defaultdict(set)
        self._events: Dict[EventCall, Set[Node]] = defaultdict(set)
        self._written: Dict[Variable, Set[Node]] = defaultdict(set)

    @property
    def send_eth(self) -> Dict[Node, Set[Node]]:
        """
        Return the list of calls sending value
        :return:
        """
        return self._send_eth

    @property
    def calls(self) -> Dict[Node, Set[Node]]:
        """
        Return the list of calls that can callback
        :return:
        """
        return self._calls

    @property
    def reads(self) -> Dict[Variable, Set[Node]]:
        """
        Return of variables that are read
        :return:
        """
        return self._reads

    @property
    def written(self) -> Dict[Variable, Set[Node]]:
        """
        Return of variables that are written
        :return:
        """
        return self._written

    @property
    def reads_prior_calls(self) -> Dict[Node, Set[Variable]]:
        """
        Return the dictionary node -> variables read before any call
        :return:
        """
        return self._reads_prior_calls

    @property
    def events(self) -> Dict[EventCall, Set[Node]]:
        """
        Return the list of events
        :return:
        """
        return self._events

    def merge_fathers(self, node, skip_father, detector):
        for father in node.fathers:
            if detector.KEY in father.context:
                self._send_eth = union_dict(
                    self._send_eth,
                    {
                        key: values
                        for key, values in father.context[detector.KEY].send_eth.items()
                        if key != skip_father
                    },
                )
                self._calls = union_dict(
                    self._calls,
                    {
                        key: values
                        for key, values in father.context[detector.KEY].calls.items()
                        if key != skip_father
                    },
                )
                self._reads = union_dict(
                    self._reads, father.context[detector.KEY].reads
                )
                self._reads_prior_calls = union_dict(
                    self.reads_prior_calls,
                    father.context[detector.KEY].reads_prior_calls,
                )

    def analyze_node(self, node, detector):
        state_vars_read: Dict[Variable, Set[Node]] = defaultdict(
            set, {v: {node} for v in node.state_variables_read}
        )

        # All the state variables written
        state_vars_written: Dict[Variable, Set[Node]] = defaultdict(
            set, {v: {node} for v in node.state_variables_written}
        )
        slithir_operations = []
        # Add the state variables written in internal calls
        for internal_call in node.internal_calls:
            # Filter to Function, as internal_call can be a solidity call
            if isinstance(internal_call, Function):
                for internal_node in internal_call.all_nodes():
                    for read in internal_node.state_variables_read:
                        state_vars_read[read].add(internal_node)
                    for write in internal_node.state_variables_written:
                        state_vars_written[write].add(internal_node)
                slithir_operations += internal_call.all_slithir_operations()

        contains_call = False

        self._written = state_vars_written
        for ir in node.irs + slithir_operations:
            if detector.can_callback(ir):
                self._calls[node] |= {ir.node}
                self._reads_prior_calls[node] = set(
                    self._reads_prior_calls.get(node, set())
                    | set(node.context[detector.KEY].reads.keys())
                    | set(state_vars_read.keys())
                )
                contains_call = True

            if detector.can_send_eth(ir):
                self._send_eth[node] |= {ir.node}

            if isinstance(ir, EventCall):
                self._events[ir] |= {ir.node, node}

        self._reads = union_dict(self._reads, state_vars_read)

        return contains_call

    def add(self, fathers):
        self._send_eth = union_dict(self._send_eth, fathers.send_eth)
        self._calls = union_dict(self._calls, fathers.calls)
        self._reads = union_dict(self._reads, fathers.reads)
        self._reads_prior_calls = union_dict(
            self._reads_prior_calls, fathers.reads_prior_calls
        )

    def does_not_bring_new_info(self, new_info):
        if is_subset(new_info.calls, self.calls):
            if is_subset(new_info.send_eth, self.send_eth):
                if is_subset(new_info.reads, self.reads):
                    if dict_are_equal(
                        new_info.reads_prior_calls, self.reads_prior_calls
                    ):
                        return True
        return False


def _filter_if(node):
    """
    Check if the node is a condtional node where
    there is an external call checked
    Heuristic:
        - The call is a IF node
        - It contains a, external call
        - The condition is the negation (!)
    This will work only on naive implementation
    """
    return (
        isinstance(node.expression, UnaryOperation)
        and node.expression.type == UnaryOperationType.BANG
    )


class Reentrancy(AbstractDetector):
    KEY = "REENTRANCY"
    WIKI_EXPLOIT_SCENARIO = "Check original reentrancy"
    WIKI_RECOMMENDATION = "Check original reentrancy"

    # can_callback and can_send_eth are static method
    # allowing inherited classes to define different behaviors
    # For example reentrancy_no_gas consider Send and Transfer as reentrant functions
    @staticmethod
    def can_callback(ir):
        """
        Detect if the node contains a call that can
        be used to re-entrance
        Consider as valid target:
        - low level call
        - high level call
        """
        return isinstance(ir, Call) and ir.can_reenter()

    @staticmethod
    def can_send_eth(ir):
        """
        Detect if the node can send eth
        """
        return isinstance(ir, Call) and ir.can_send_eth()

    def _explore(self, node, visited, skip_father=None):
        """
        Explore the CFG and look for re-entrancy
        Heuristic: There is a re-entrancy if a state variable is written
                    after an external call
        node.context will contains the external calls executed
        It contains the calls executed in father nodes
        if node.context is not empty, and variables are written, a re-entrancy is possible
        """
        if node in visited:
            return

        visited = visited + [node]

        fathers_context = AbstractState()
        fathers_context.merge_fathers(node, skip_father, self)

        # Exclude path that dont bring further information
        if node in self.visited_all_paths:
            if self.visited_all_paths[node].does_not_bring_new_info(fathers_context):
                return
        else:
            self.visited_all_paths[node] = AbstractState()

        self.visited_all_paths[node].add(fathers_context)

        node.context[self.KEY] = fathers_context

        contains_call = fathers_context.analyze_node(node, self)
        node.context[self.KEY] = fathers_context

        sons = node.sons
        if contains_call and node.type in [NodeType.IF, NodeType.IFLOOP]:
            if _filter_if(node):
                son = sons[0]
                self._explore(son, visited, node)
                sons = sons[1:]
            else:
                son = sons[1]
                self._explore(son, visited, node)
                sons = [sons[0]]

        for son in sons:
            self._explore(son, visited)

    def detect_reentrancy(self, contract):
        for function in contract.functions_and_modifiers_declared:
            if not function.is_constructor:
                if function.is_implemented:
                    if self.KEY in function.context:
                        continue
                    self._explore(function.entry_point, [])
                    function.context[self.KEY] = True

    def _detect(self):
        """"""
        # if a node was already visited by another path
        # we will only explore it if the traversal brings
        # new variables written
        # This speedup the exploration through a light fixpoint
        # Its particular useful on 'complex' functions with several loops and conditions
        self.visited_all_paths = {}  # pylint: disable=attribute-defined-outside-init

        for c in self.contracts:
            self.detect_reentrancy(c)

        return []



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification


class BeforeTokenTransfer(AbstractDetector):
    """
    Sees if a contract has a beforeTokenTransfer function.
    """

    ARGUMENT = 'pess-before-token-transfer' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'beforeTokenTransfer function does not follow OZ documentation'
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = 'https://docs.openzeppelin.com/contracts/4.x/extending-contracts#rules_of_hooks'
    WIKI_TITLE = 'Before Token Transfer'
    WIKI_DESCRIPTION = "Follow OZ documentation using their contracts"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Make sure that beforeTokenTransfer function is used in the correct way.'


    # def _has_no_virtual (self, fun: Function) -> bool: 
    #     """Checks if function does not have virtual modifier"""
    #     if isinstance(fun, Function):   # check for a correct function type
    #         print(fun.source_mapping) # reads function modifiers
    #     return False

    # def _has_no_super (self, fun: Function) -> bool:
    #     """Checks if parent function call has super"""


    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions_and_modifiers_declared:
                # x = self._has_no_virtual(f) TODO: add
                # y = self._has_no_super(f) TODO: add
                if "beforeTokenTransfer" in f.name:
                    res.append(self.generate_result([
                        "beforeTokenTransfer in ",
                        f, ' must have virtual and super.',
                        '\n']))
        return res



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function


class TimelockController(AbstractDetector):
    """
    Sees if contract uses OZ TimelockController
    """

    ARGUMENT = 'pess-timelock-controller' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'The OZ implementation of timelock controller is used'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/timelock_controller.md'
    WIKI_TITLE = 'OZ Timelock Controller'
    WIKI_DESCRIPTION = "Delpoyer can bypass timelock limitations"
    WIKI_EXPLOIT_SCENARIO = '-'
    WIKI_RECOMMENDATION = 'Revoke Proposer and Executor roles from the deployer of OZ Timelock Controller'


    def _function_timelock_controller (self, fun: Function) -> bool: 
        """Checks if function has multiple storage read"""
        if isinstance(fun, Function):   # check for a correct function type
            for parameter in fun.parameters:
                if str(parameter.type) == "TimelockController":
                    return True
            for n in fun.nodes: # check every node for a TimelockController
                if "TimelockController" in str(n):
                    return True
        return False

    def _state_var_timelock_controller (self, contract) -> bool:
        """Checks if state var contatins TimelockController"""
        for state_var in contract.state_variables:
            if str(state_var.type) == "TimelockController":
                return True
        return False

    def _inheritance_timelock_controller (self, contract) -> bool:
        """Check if TimelockController contract is inherited"""
        for inherited_contract in contract.inheritance:
            if str(inherited_contract.name) == "TimelockController":
                return True
        return False

    def _detect(self) -> List[Output]:
        """Main function"""
        res = []
        for contract in self.compilation_unit.contracts_derived:
            found_in_inheritance = self._inheritance_timelock_controller(contract)
            found_in_state_var = self._state_var_timelock_controller(contract)
            for f in contract.functions_and_modifiers_declared:
                found_in_function = self._function_timelock_controller(f)
                if found_in_state_var or found_in_function or found_in_inheritance:
                    res.append(self.generate_result([
                        "In a contract", ' ',
                        contract, ' timelock-controller is used. Revoke deployer roles.'
                        '\n']))
        return res



import re
from typing import List, Tuple

from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import (
    SolidityCall,
    Condition,
)
from slither.core.declarations import (
    FunctionContract,
)
from slither.core.cfg.node import Node, NodeType
from slither.slithir.operations import LowLevelCall
from slither.analyses.data_dependency.data_dependency import is_dependent


class Ecrecover(AbstractDetector):
    """
    Detects not checked results of ecrecover
    """

    ARGUMENT = "pess-ecrecover"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "signer = ecrecover(hash, v, r, s)"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/ecrecover.md"
    WIKI_TITLE = "Ecrecover"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = "Attacker can validate signatures from 0x0 address"
    WIKI_RECOMMENDATION = "Check the result of ecrecover"

    def analyze_function(
        self, function: FunctionContract
    ) -> List[Tuple[FunctionContract, Node, LowLevelCall, bool, bool]]:
        unchecked_results = set()
        var_to_node = {}
        for node in function.nodes:
            for ir in node.irs:
                try:
                    node_contains_0 = re.search(
                        r"address\((0|0x0*)\)", str(node)
                    )  # check if the node contains address(0|0x0..)
                    if isinstance(ir, SolidityCall):
                        if (
                            ir.function.name
                            == "ecrecover(bytes32,uint8,bytes32,bytes32)"
                        ):
                            unchecked_results.add(ir.lvalue)
                            # print(ir.lvalue, node)
                            var_to_node[ir.lvalue] = node
                        elif (
                            ir.function.name == "require(bool,string)"
                            or ir.function.name == "assert(bool)"
                        ):
                            if not node_contains_0:  # does not contain 0 check
                                continue
                            checking_var = ir.arguments[0]
                            for ur in unchecked_results:
                                if is_dependent(checking_var, ur, node):
                                    unchecked_results.remove(ur)
                                    break
                    elif isinstance(ir, Condition):
                        # this is copypaste, for now, couldn't figure out how to make this better without overcomplicating
                        if not node_contains_0:  # does not contain 0 check
                            continue
                        for ur in unchecked_results:
                            if is_dependent(ir.value, ur, node):
                                unchecked_results.remove(ur)
                                break

                except Exception as e:
                    print("failed", e)
        
        for node in function.nodes:
            if node._node_type == NodeType.RETURN:
                node_contains_0 = re.search(
                        r"address\((0|0x0*)\)", str(node)
                    )  # check if the node contains address(0|0x0..)
                if not node_contains_0:
                    continue
                temp_unchecked = set(unchecked_results)
                for ur in temp_unchecked:
                    if is_dependent(ur, ur, node):
                        unchecked_results.remove(ur)

        return [var_to_node[ur] for ur in unchecked_results]

    def _detect(self):
        results = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                res = self.analyze_function(f)
                if res:
                    for r in res:
                        info = ["Unchecked result of ecrecover for 0:\n\t", r, "\n"]
                        tres = self.generate_result(info)
                        tres.add(r)
                        results.append(tres)
        return results



from typing import List, Tuple
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import TypeConversion, Operation
from slither.core.declarations import FunctionContract
from slither.core.cfg.node import Node


def is_ok_cast(from_type: str, to_type: str) -> bool:
    if from_type == to_type:
        return True
    if from_type == "address" or to_type == "address":
        # address <-> interface/contract/this
        # address <-> uint160 (solidity doesn't allow other direct casts of address)
        return True

    if "bytes" in from_type and "bytes" in to_type:
        # bytesX <-> bytesY will have different results
        return False

    if "bytes" in from_type or "bytes" in to_type:
        return True

    if "uint" == from_type[:4] and "uint" in to_type[:4]:
        size_from = from_type[4:]
        if size_from == "":
            size_from = "256"

        size_to = to_type[4:]
        if to_type == "":
            size_to = "256"

        if int(size_from) > int(size_to):
            return False
        return True

    if "int" == from_type[:3] and "int" == to_type[:3]:
        size_from = from_type[3:]
        if size_from == "":
            size_from = "256"

        size_to = to_type[3:]
        if to_type == "":
            size_to = "256"

        if int(size_from) > int(size_to):
            return False
        return True

    sorted_types = sorted([from_type, to_type])
    if "int" == sorted_types[0][:3] and "uint" == sorted_types[1][:4]:
        return False

    return True


class DubiousTypecast(AbstractDetector):
    """
    Shows nonstandard typecasts.
    """

    ARGUMENT = "pess-dubious-typecast"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "uint8 = uint8(uint256)"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/dubious_typecast.md"
    WIKI_TITLE = "Dubious Typecast"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = (
        "Can produce unpredictable results because of nonstandard typecasts"
    )
    WIKI_RECOMMENDATION = "Use clear constants"

    def analyze_irs(self, irs: List[Operation]) -> List[Tuple[str, str]]:
        results = []
        for i in irs:
            try:
                if isinstance(i, TypeConversion):
                    from_type = i.variable.type
                    to_type = i.type

                    if not (is_ok_cast(str(from_type), str(to_type))):
                        results.append((from_type, to_type))
            except Exception as e:
                print("DubiousTypecast:Failed to check types", e)
                print(i)

        return results

    def get_dubious_typecasts(self, fun: FunctionContract, params=None):
        results: List[Tuple[Node, List[Tuple[str, str]]]] = []
        for node in fun.nodes:
            node_res = self.analyze_irs(node.irs)
            if node_res:
                results.append((node, node_res))
        return results

    def _detect(self):
        results = []
        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                func_res = self.get_dubious_typecasts(f)
                if func_res:
                    info = ["Dubious typecast in ", f, ":\n"]
                    for node, node_res in func_res:
                        for from_type, to_type in node_res:
                            info += [
                                f"\t{str(from_type)} => {str(to_type)} casting occurs in ",
                                node,
                                "\n",
                            ]
                    res = self.generate_result(info)
                    res.add(f)
                    results.append(res)

        return results



from slitherin.detectors.arbitrary_call.arbitrary_call import ArbitraryCall
from slitherin.detectors.double_entry_token_possibility import (
    DoubleEntryTokenPossiblity,
)
from slitherin.detectors.dubious_typecast import DubiousTypecast
from slitherin.detectors.falsy_only_eoa_modifier import OnlyEOACheck
from slitherin.detectors.magic_number import MagicNumber
from slitherin.detectors.strange_setter import StrangeSetter
from slitherin.detectors.unprotected_setter import UnprotectedSetter
from slitherin.detectors.nft_approve_warning import NftApproveWarning
from slitherin.detectors.inconsistent_nonreentrant import InconsistentNonreentrant
from slitherin.detectors.call_forward_to_protected import CallForwardToProtected
from slitherin.detectors.multiple_storage_read import MultipleStorageRead
from slitherin.detectors.timelock_controller import TimelockController
from slitherin.detectors.tx_gasprice_warning import TxGaspriceWarning
from slitherin.detectors.unprotected_initialize import UnprotectedInitialize
from slitherin.detectors.read_only_reentrancy import ReadOnlyReentrancy
from slitherin.detectors.event_setter import EventSetter
from slitherin.detectors.before_token_transfer import BeforeTokenTransfer
from slitherin.detectors.uni_v2 import UniswapV2
from slitherin.detectors.token_fallback import TokenFallback
from slitherin.detectors.for_continue_increment import ForContinueIncrement
from slitherin.detectors.ecrecover import Ecrecover
from slitherin.detectors.public_vs_external import PublicVsExternal
from slitherin.detectors.aave.flashloan_callback import AAVEFlashloanCallbackDetector
from slitherin.detectors.arbitrum.arbitrum_prevrandao_difficulty import (
    ArbitrumPrevrandaoDifficulty,
)
from slitherin.detectors.arbitrum.block_number_timestamp import (
    ArbitrumBlockNumberTimestamp,
)
from slitherin.detectors.potential_arith_overflow import PotentialArithmOverflow

artbitrum_detectors = [
    ArbitrumPrevrandaoDifficulty,
    ArbitrumBlockNumberTimestamp,
]

plugin_detectors = artbitrum_detectors + [
    DoubleEntryTokenPossiblity,
    UnprotectedSetter,
    NftApproveWarning,
    InconsistentNonreentrant,
    StrangeSetter,
    OnlyEOACheck,
    MagicNumber,
    DubiousTypecast,
    CallForwardToProtected,
    MultipleStorageRead,
    TimelockController,
    TxGaspriceWarning,
    UnprotectedInitialize,
    ReadOnlyReentrancy,
    EventSetter,
    BeforeTokenTransfer,
    UniswapV2,
    TokenFallback,
    ForContinueIncrement,
    ArbitraryCall,
    Ecrecover,
    PublicVsExternal,
    AAVEFlashloanCallbackDetector,
    PotentialArithmOverflow
]
plugin_printers = []


def make_plugin():
    return plugin_detectors, plugin_printers



from setuptools import setup, find_packages
from slitherin.cli import SLITHERIN_VERSION

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="slitherin",
    description="Pessimistic security Slither detectors",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={"slitherin": ["py.typed"]},
    url="https://github.com/pessimistic-io/slitherin",
    author="Pessimistic.io",
    author_email="info@pessimistic.io",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries",
    ],
    version=SLITHERIN_VERSION,
    packages=find_packages(),
    license="AGPL-3.0",
    python_requires=">=3.8",
    install_requires=["slither-analyzer>=0.10.0"],
    extras_requires={
        "dev": ["twine>=4.0.2"],
    },
    entry_points={
        "slither_analyzer.plugin": "slither slitherin-plugins=slitherin:make_plugin",
        "console_scripts": ["slitherin=slitherin.cli:main"],
    },
    include_package_data=True,
)



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Contract


class PublicVsExternal(AbstractDetector):
    """
    Detects functions that have "public" modifiers and not used in the contract
    """

    ARGUMENT = "pess-public-vs-external"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "mark public functions as external where possible, to enhance contract's control-flow readability"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/public_vs_external.md"
    WIKI_TITLE = "Public vs External"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = "No exploits, just readability enhancement"
    WIKI_RECOMMENDATION = "Mark public functions as external where it is possible"

    def _analyze_contract(self, contract: Contract) -> list:
        res = []
        used_functions = set()
        for f in contract.functions_and_modifiers_declared:
            for node in f.nodes:
                for call in node.internal_calls:
                    used_functions.add(call.name)

        for f in contract.functions_and_modifiers_declared:
            if f.visibility == "public" and f.name not in used_functions and not f.is_constructor:
                res.append(f)
        return res

    def _detect(self) -> List[Output]:
        """Main function"""
        results = []
        for contract in self.compilation_unit.contracts_derived:
            res = self._analyze_contract(contract)
            if res:
                info = [
                    "The following public functions could be turned into external in ",
                    contract,
                    " contract:\n",
                ]
                for r in res:
                    info += ["\t", r, "\n"]
                contract_result = self.generate_result(info)
                results.append(contract_result)

        return results
